# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import importlib
import inspect
import os.path
import pkgutil
import subprocess
import sys

import uproot

order = [
    "uproot",
    "uproot.reading",
    "uproot.writing",
    "uproot.behaviors",
    "uproot._dask",
    "uproot.behavior",
    "uproot.model",
    "uproot.streamers",
    "uproot.cache",
    "uproot.compression",
    "uproot.deserialization",
    "uproot.serialization",
    "uproot.pyroot",
    "uproot.source",
    "uproot.sink",
    "uproot.interpretation",
    "uproot.containers",
    "uproot.language",
    "uproot.models",
    "uproot.exceptions",
]

common = [
    "uproot.reading.open",
    "uproot.behaviors.TBranch.iterate",
    "uproot.behaviors.TBranch.concatenate",
    "uproot._dask.dask",
    "uproot.writing.writable.create",
    "uproot.writing.writable.recreate",
    "uproot.writing.writable.update",
    "uproot.reading.ReadOnlyFile",
    "uproot.reading.ReadOnlyDirectory",
    "uproot.behaviors.TTree.TTree",
    "uproot.behaviors.TBranch.TBranch",
    "uproot.writing.writable.WritableFile",
    "uproot.writing.writable.WritableDirectory",
    "uproot.writing.writable.WritableTree",
    "uproot.writing.writable.WritableBranch",
    "uproot.compression.ZLIB",
    "uproot.compression.LZMA",
    "uproot.compression.LZ4",
    "uproot.compression.ZSTD",
    "uproot.cache.LRUCache",
    "uproot.cache.LRUArrayCache",
    "uproot.model.Model",
    "uproot.pyroot.from_pyroot",
    "uproot.source.object.ObjectSource",
    "uproot.source.file.MemmapSource",
    "uproot.source.file.MultithreadedFileSource",
    "uproot.source.http.HTTPSource",
    "uproot.source.http.MultithreadedHTTPSource",
    "uproot.source.xrootd.XRootDSource",
    "uproot.source.xrootd.MultithreadedXRootDSource",
    "uproot.models.TTree.num_entries",
]

latest_commit = (
    subprocess.run(["git", "rev-parse", "HEAD"], stdout=subprocess.PIPE)
    .stdout.decode("utf-8")
    .strip()
)

main = open("main.toctree", "w")
main.write(
    """.. toctree::
    :caption: Main Interface
    :hidden:

{}""".format(
        "".join(f"    {x}\n" for x in common)
    )
)
toctree = open("uproot.toctree", "w")
toctree.write(
    """.. toctree::
    :caption: Detailed Reference
    :hidden:

"""
)
toctree2 = None


def ensure(filename, content):
    overwrite = not os.path.exists(filename)
    if not overwrite:
        overwrite = open(filename).read() != content
    if overwrite:
        open(filename, "w").write(content)
        sys.stderr.write(filename + " (OVERWRITTEN)\n")
    else:
        sys.stderr.write(filename + "\n")


def handle_module(modulename, module):
    if any(x.startswith("_") for x in modulename.split(".")) and not any(
        x == "_dask" for x in modulename.split(".")
    ):
        return

    content = """{0}
{1}

.. automodule:: {0}
""".format(
        modulename, "=" * len(modulename)
    )
    ensure(modulename + ".rst", content)
    if toctree2 is None:
        toctree.write("    " + modulename + " (module) <" + modulename + ">\n")
    else:
        toctree2.write("    " + modulename + " (module) <" + modulename + ">\n")

    if modulename != "uproot" and all(
        not x.startswith("_") or x == "_dask" for x in modulename.split(".")
    ):

        def good(obj):
            if inspect.isfunction(obj) or inspect.isclass(obj):
                if obj.__module__ == modulename:
                    return True
            return False

        def line_order(pair):
            name, obj = pair
            try:
                return inspect.getsourcelines(obj)[1]
            except OSError:
                return float("inf")

        for pair in sorted(inspect.getmembers(module, good), key=line_order):
            name, obj = pair
            if not name.startswith("_"):
                if inspect.isclass(obj):
                    handle_class(modulename + "." + name, obj)
                elif inspect.isfunction(obj):
                    handle_function(modulename + "." + name, obj)


def handle_class(classname, cls):
    def line_order(obj):
        if isinstance(obj, property):
            obj = obj.fget
        return inspect.getsourcelines(obj)[1]

    methods = {}
    mro = list(cls.__mro__)

    if hasattr(uproot, cls.__name__):
        title = "uproot." + cls.__name__
        upfront = True
    else:
        title = classname
        upfront = False

    for index, basecls in enumerate(mro):
        if basecls.__module__.startswith("uproot."):

            def good(obj):
                if inspect.ismethod(obj) or inspect.isfunction(obj):
                    module, name = obj.__module__, obj.__name__
                elif isinstance(obj, property):
                    try:
                        module, name = obj.fget.__module__, obj.fget.__name__
                    except AttributeError:
                        return False
                else:
                    module, name = "", ""
                if module.startswith("uproot."):
                    if index + 1 >= len(mro) or obj is not getattr(
                        mro[index + 1], name, None
                    ):
                        return True
                return False

            for name, obj in inspect.getmembers(basecls, good):
                if name in basecls.__dict__ and not name.startswith("_"):
                    fill = []
                    fill.append(".. _" + classname + "." + name + ":" + "\n")
                    fill.append(name)
                    fill.append("-" * len(fill[-1]))
                    fill.append("")
                    if basecls is not cls:
                        fill.append(
                            "Inherited from :doc:`{}`.".format(
                                basecls.__module__ + "." + basecls.__name__
                            )
                        )
                        fill.append("")
                    if isinstance(obj, property):
                        fill.append(".. autoattribute:: " + classname + "." + name)
                    elif callable(obj):
                        fill.append(".. automethod:: " + classname + "." + name)
                    else:
                        fill.append(".. autoattribute:: " + classname + "." + name)
                    fill.append("")
                    methods[name] = (index, line_order(obj), "\n".join(fill))

    def prettymro(c):
        fullname = c.__module__ + "." + c.__name__
        if c.__module__.startswith("uproot."):
            return "#. :doc:`" + fullname + "`"
        else:
            return "#. ``" + fullname + "``"

    fullfilename = importlib.import_module(cls.__module__).__file__
    shortfilename = fullfilename[fullfilename.rindex("uproot/") :]
    link = "`{} <https://github.com/scikit-hep/uproot4/blob/{}/src/{}>`__".format(
        cls.__module__, latest_commit, shortfilename
    )
    try:
        linelink = "`line {0} <https://github.com/scikit-hep/uproot4/blob/{1}/src/{2}#L{0}>`__".format(
            inspect.getsourcelines(cls)[1], latest_commit, shortfilename
        )
    except OSError:
        linelink = ""

    inheritance_header = ""
    inheritance_footer = ""
    inheritance_sep = ""
    inheritance = [prettymro(c) for c in cls.__mro__[1:] if c is not object]
    if len(inheritance) > 0:
        longest_cell = max(max(len(x) for x in inheritance), 22)
        inheritance_header = """.. table::
    :class: note

    +-{}-+
    | **Inheritance order:** {}|
    +={}=+
    | """.format(
            "-" * longest_cell, " " * (longest_cell - 22), "=" * longest_cell
        )
        inheritance_footer = """ |
    +-{}-+""".format(
            "-" * longest_cell
        )
        inheritance = [x + " " * (longest_cell - len(x)) for x in inheritance]
        inheritance_sep = """ |
    | """

    content = """{}
{}

Defined in {} on {}.

{}

.. autoclass:: {}

{}
""".format(
        title,
        "=" * len(title),
        link,
        linelink,
        inheritance_header + inheritance_sep.join(inheritance) + inheritance_footer,
        classname,
        "\n".join([text for index, line, text in sorted(methods.values())]),
    )

    ensure(classname + ".rst", content)
    if upfront or toctree2 is None:
        if classname not in common:
            toctree.write("    " + classname + "\n")
        toctree2.write("    " + classname + " <" + classname + ">\n")
    else:
        toctree2.write("    " + classname + "\n")


def handle_function(functionname, cls):
    if hasattr(uproot, cls.__name__):
        title = "uproot." + cls.__name__
        upfront = True
    else:
        title = functionname
        upfront = False

    fullfilename = importlib.import_module(cls.__module__).__file__
    shortfilename = fullfilename[fullfilename.rindex("uproot/") :]
    link = "`{} <https://github.com/scikit-hep/uproot4/blob/{}/src/{}>`__".format(
        cls.__module__, latest_commit, shortfilename
    )
    linelink = "`line {0} <https://github.com/scikit-hep/uproot4/blob/{1}/src/{2}#L{0}>`__".format(
        inspect.getsourcelines(cls)[1], latest_commit, shortfilename
    )

    content = """{}
{}

Defined in {} on {}.

.. autofunction:: {}
""".format(
        title, "=" * len(title), link, linelink, functionname
    )
    ensure(functionname + ".rst", content)
    if upfront or toctree2 is None:
        if functionname not in common:
            toctree.write("    " + functionname + "\n")
        toctree2.write("    " + functionname + " <" + functionname + ">\n")
    else:
        toctree2.write("    " + functionname + "\n")


for modulename in order:
    module = importlib.import_module(modulename)

    if modulename != "uproot":
        toctree2 = open(modulename + ".toctree", "w")
        toctree2.write(
            """.. toctree::
    :caption: {}
    :hidden:

""".format(
                modulename.replace("uproot.", "")
            )
        )

    handle_module(modulename, module)
    if module.__file__.endswith("__init__.py") and modulename != "uproot":
        for submodulename in sorted(
            modulename + "." + name
            for loader, name, is_pkg in pkgutil.walk_packages(module.__path__)
        ):
            submodule = importlib.import_module(submodulename)
            handle_module(submodulename, submodule)

toctree.close()
toctree2.close()



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import numpy as np
import pytest
import skhep_testdata

import uproot

ROOT = pytest.importorskip("ROOT")


def test(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    with uproot.recreate(filename) as file:
        file["\u03c0"] = np.histogram(np.random.normal(0, 1, 10))
        assert file["\u03c0"].name == "\u03c0"

    file2 = ROOT.TFile(filename)
    histogram = file2.Get("\u03c0")
    assert histogram.GetName() == "\u03c0"



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import numpy as np
import pytest
import skhep_testdata

import uproot


def test():
    with uproot.open(skhep_testdata.data_path("uproot-mc10events.root")) as file:
        tree = file["Events"]
        assert tree["Muon"].array(library="np").tolist() == [
            1,
            0,
            3,
            3,
            0,
            1,
            5,
            0,
            1,
            0,
        ]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy
import pytest
import skhep_testdata

import uproot


def test():
    path = skhep_testdata.data_path("uproot-HZZ-objects.root")
    with uproot.open(path)["events"] as tree:
        form = tree["jetp4"].interpretation.awkward_form(None)
        assert form.to_dict() == {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "RecordArray",
                "fields": ["fP", "fE"],
                "contents": [
                    {
                        "class": "RecordArray",
                        "fields": ["fX", "fY", "fZ"],
                        "contents": [
                            {
                                "class": "NumpyArray",
                                "primitive": "float64",
                                "inner_shape": [],
                                "parameters": {},
                                "form_key": None,
                            },
                            {
                                "class": "NumpyArray",
                                "primitive": "float64",
                                "inner_shape": [],
                                "parameters": {},
                                "form_key": None,
                            },
                            {
                                "class": "NumpyArray",
                                "primitive": "float64",
                                "inner_shape": [],
                                "parameters": {},
                                "form_key": None,
                            },
                        ],
                        "parameters": {"__record__": "TVector3"},
                        "form_key": None,
                    },
                    {
                        "class": "NumpyArray",
                        "primitive": "float64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": None,
                    },
                ],
                "parameters": {"__record__": "TLorentzVector"},
                "form_key": None,
            },
            "parameters": {},
            "form_key": None,
        }



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys

import numpy
import pytest
import skhep_testdata

import uproot

from_ROOT = {}

from_ROOT["one"] = json.loads(
    r"""{
  "_typename" : "TH1F",
  "fUniqueID" : 0,
  "fBits" : 8,
  "fName" : "one",
  "fTitle" : "numero uno",
  "fLineColor" : 602,
  "fLineStyle" : 1,
  "fLineWidth" : 1,
  "fFillColor" : 0,
  "fFillStyle" : 1001,
  "fMarkerColor" : 1,
  "fMarkerStyle" : 1,
  "fMarkerSize" : 1.0,
  "fNcells" : 12,
  "fXaxis" : {
    "_typename" : "TAxis",
    "fUniqueID" : 0,
    "fBits" : 0,
    "fName" : "xaxis",
    "fTitle" : "",
    "fNdivisions" : 510,
    "fAxisColor" : 1,
    "fLabelColor" : 1,
    "fLabelFont" : 42,
    "fLabelOffset" : 0.004999999888241291,
    "fLabelSize" : 0.03500000014901161,
    "fTickLength" : 0.029999999329447746,
    "fTitleOffset" : 1.0,
    "fTitleSize" : 0.03500000014901161,
    "fTitleColor" : 1,
    "fTitleFont" : 42,
    "fNbins" : 10,
    "fXmin" : -3.0,
    "fXmax" : 3.0,
    "fXbins" : [],
    "fFirst" : 0,
    "fLast" : 0,
    "fBits2" : 0,
    "fTimeDisplay" : false,
    "fTimeFormat" : "",
    "fLabels" : null,
    "fModLabs" : null
  },
  "fYaxis" : {
    "_typename" : "TAxis",
    "fUniqueID" : 0,
    "fBits" : 0,
    "fName" : "yaxis",
    "fTitle" : "",
    "fNdivisions" : 510,
    "fAxisColor" : 1,
    "fLabelColor" : 1,
    "fLabelFont" : 42,
    "fLabelOffset" : 0.004999999888241291,
    "fLabelSize" : 0.03500000014901161,
    "fTickLength" : 0.029999999329447746,
    "fTitleOffset" : 1.0,
    "fTitleSize" : 0.03500000014901161,
    "fTitleColor" : 1,
    "fTitleFont" : 42,
    "fNbins" : 1,
    "fXmin" : 0.0,
    "fXmax" : 1.0,
    "fXbins" : [],
    "fFirst" : 0,
    "fLast" : 0,
    "fBits2" : 0,
    "fTimeDisplay" : false,
    "fTimeFormat" : "",
    "fLabels" : null,
    "fModLabs" : null
  },
  "fZaxis" : {
    "_typename" : "TAxis",
    "fUniqueID" : 0,
    "fBits" : 0,
    "fName" : "zaxis",
    "fTitle" : "",
    "fNdivisions" : 510,
    "fAxisColor" : 1,
    "fLabelColor" : 1,
    "fLabelFont" : 42,
    "fLabelOffset" : 0.004999999888241291,
    "fLabelSize" : 0.03500000014901161,
    "fTickLength" : 0.029999999329447746,
    "fTitleOffset" : 1.0,
    "fTitleSize" : 0.03500000014901161,
    "fTitleColor" : 1,
    "fTitleFont" : 42,
    "fNbins" : 1,
    "fXmin" : 0.0,
    "fXmax" : 1.0,
    "fXbins" : [],
    "fFirst" : 0,
    "fLast" : 0,
    "fBits2" : 0,
    "fTimeDisplay" : false,
    "fTimeFormat" : "",
    "fLabels" : null,
    "fModLabs" : null
  },
  "fBarOffset" : 0,
  "fBarWidth" : 1000,
  "fEntries" : 10000.0,
  "fTsumw" : 10000.0,
  "fTsumw2" : 10000.0,
  "fTsumwx" : 81.87497264376279,
  "fTsumwx2" : 10388.152621259549,
  "fMaximum" : -1111.0,
  "fMinimum" : -1111.0,
  "fNormFactor" : 0.0,
  "fContour" : [],
  "fSumw2" : [],
  "fOption" : "",
  "fFunctions" : {
    "_typename" : "TList",
    "name" : "TList",
    "arr" : [],
    "opt" : []
  },
  "fBufferSize" : 0,
  "fBuffer" : [],
  "fBinStatErrOpt" : 0,
  "fStatOverflows" : 2,
  "fArray" : [0, 68, 285, 755, 1580, 2296, 2286, 1570, 795, 289, 76, 0]
}"""
)


def drop_fbits(x):
    if isinstance(x, dict):
        return {
            k: drop_fbits(v)
            for k, v in x.items()
            if k not in ("fBits", "fStatOverflows", "fN", "fArray")
        }
    elif isinstance(x, list):
        return list(drop_fbits(v) for v in x)
    else:
        return x


def test():
    with uproot.open(skhep_testdata.data_path("uproot-histograms.root")) as f:
        assert f.file.class_named("TH1", 7).member_names == [
            "fNcells",
            "fXaxis",
            "fYaxis",
            "fZaxis",
            "fBarOffset",
            "fBarWidth",
            "fEntries",
            "fTsumw",
            "fTsumw2",
            "fTsumwx",
            "fTsumwx2",
            "fMaximum",
            "fMinimum",
            "fNormFactor",
            "fContour",
            "fSumw2",
            "fOption",
            "fFunctions",
            "fBufferSize",
            "fBuffer",
            "fBinStatErrOpt",
        ]

        assert drop_fbits(from_ROOT["one"]) == drop_fbits(f.get("one").tojson())



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import pytest

import uproot
import uproot.writing

ROOT = pytest.importorskip("ROOT")


def test_recreate(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    with uproot.recreate(filename) as f1:
        f1["hey"] = "you"
        f1["subdir/there"] = "you guys"

    with uproot.open(filename) as f2:
        assert f2["hey"] == "you"
        assert f2["subdir/there"] == "you guys"
        assert list(f2.file.streamers) == ["TObjString"]

    f3 = ROOT.TFile(filename, "update")
    assert [x.GetName() for x in f3.GetStreamerInfoList()] == ["TObjString"]
    assert f3.Get("hey") == "you"
    assert f3.Get("subdir/there") == "you guys"
    f3.cd("subdir")
    x = ROOT.TObjString("wowie")
    x.Write()
    f3.Close()

    with uproot.open(filename) as f4:
        assert f4["hey"] == "you"
        assert f4["subdir/there"] == "you guys"
        assert f4["subdir/wowie"] == "wowie"
        assert list(f4.file.streamers) == ["TObjString"]


def test_update(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    f1 = ROOT.TFile(filename, "recreate")
    f1.mkdir("subdir")
    f1.cd("subdir")
    x = ROOT.TObjString("wowie")
    x.Write()
    f1.Close()

    with uproot.open(filename) as f2:
        assert f2["subdir/wowie"] == "wowie"
        assert list(f2.file.streamers) == ["TObjString"]

    with uproot.update(filename) as f3:
        f3["hey"] = "you"
        f3["subdir/there"] = "you guys"

    with uproot.open(filename) as f4:
        assert f4["hey"] == "you"
        assert f4["subdir/there"] == "you guys"
        assert f4["subdir/wowie"] == "wowie"
        assert list(f4.file.streamers) == ["TObjString"]

    f5 = ROOT.TFile(filename, "update")
    assert [x.GetName() for x in f5.GetStreamerInfoList()] == ["TObjString"]
    assert f5.Get("hey") == "you"
    assert f5.Get("subdir/there") == "you guys"
    f5.cd("subdir")
    y = ROOT.TObjString("zowie")
    y.Write()
    f5.Close()

    with uproot.open(filename) as f6:
        assert f6["hey"] == "you"
        assert f6["subdir/there"] == "you guys"
        assert f6["subdir/wowie"] == "wowie"
        assert f6["subdir/zowie"] == "zowie"
        assert list(f6.file.streamers) == ["TObjString"]


def test_serialization(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    string = "hey"
    tobjstring = uproot.writing.identify.to_TObjString(string)
    assert (
        tobjstring.tojson()["_typename"] == "TObjString"
    )  # https://github.com/scikit-hep/uproot5/issues/762
    assert tobjstring.tojson()["fString"] == str(tobjstring)

    with uproot.recreate(filename) as f1:
        f1["first"] = tobjstring
        f1["second"] = str(tobjstring)  # also checks conversion to "str"

    with uproot.open(filename) as f2:
        assert f2["first"] == f2["second"]
        assert str(f2["first"]) == string



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines utilities for modeling C++ objects as Python objects and the
:doc:`uproot.model.Model` class, which is the superclass of all objects that
are read from ROOT files.

The :doc:`uproot.model.VersionedModel` class is the superclass of all models
whose deserialization routines are specialized by ROOT class version.

A :doc:`uproot.model.DispatchByVersion` subclass selects a versioned model
after reading its version bytes.

The :doc:`uproot.model.UnknownClass` and
:doc:`uproot.model.UnknownClassVersion` are placeholders for data that could
not be modeled, either because the class has no streamer or no streamer for its
version.
"""


import re
import sys
import weakref

import numpy

import uproot

bootstrap_classnames = [
    "TStreamerInfo",
    "TStreamerElement",
    "TStreamerArtificial",
    "TStreamerBase",
    "TStreamerBasicPointer",
    "TStreamerBasicType",
    "TStreamerLoop",
    "TStreamerObject",
    "TStreamerObjectAny",
    "TStreamerObjectAnyPointer",
    "TStreamerObjectPointer",
    "TStreamerSTL",
    "TStreamerSTLstring",
    "TStreamerString",
    "TList",
    "TObjArray",
    "TClonesArray",
    "TObjString",
]
never_from_streamers = [
    "TString",
    "TList",
]

np_uint8 = numpy.dtype("u1")


def bootstrap_classes():
    """
    Returns the basic classes that are needed to load other classes (streamers,
    TList, TObjArray, TObjString).
    """
    import uproot.models.TList
    import uproot.models.TObjArray
    import uproot.models.TObjString
    import uproot.streamers

    custom_classes = {}
    for classname in bootstrap_classnames:
        custom_classes[classname] = uproot.classes[classname]

    return custom_classes


def reset_classes():
    """
    Removes all classes from ``uproot.classes`` and ``uproot.unknown_classes``
    and refills ``uproot.classes`` with original versions of these classes.
    """
    from importlib import reload

    uproot.classes = {}
    uproot.unknown_classes = {}

    reload(uproot.streamers)
    reload(uproot.models.TObject)
    reload(uproot.models.TString)
    reload(uproot.models.TArray)
    reload(uproot.models.TNamed)
    reload(uproot.models.TList)
    reload(uproot.models.THashList)
    reload(uproot.models.TObjArray)
    reload(uproot.models.TObjString)
    reload(uproot.models.TAtt)
    reload(uproot.models.TDatime)
    reload(uproot.models.TRef)

    reload(uproot.models.TTable)
    reload(uproot.models.TTree)
    reload(uproot.models.TBranch)
    reload(uproot.models.TLeaf)
    reload(uproot.models.TBasket)
    reload(uproot.models.RNTuple)
    reload(uproot.models.TH)
    reload(uproot.models.TGraph)
    reload(uproot.models.TMatrixT)


_root_alias_to_c_primitive = {
    "Bool_t": "bool",
    "Char_t": "char",
    "UChar_t": "unsigned char",
    "Short_t": "short",
    "UShort_t": "unsigned short",
    "Int_t": "int",
    "UInt_t": "unsigned int",
    "Long_t": "long",
    "ULong_t": "unsigned long",
    "Long64_t": "long long",
    "ULong64_t": "unsigned long long",
    "Size_t": "size_t",
    "Float_t": "float",
    "Double_t": "double",
    "LongDouble_t": "long double",
}

_classname_regularize = re.compile(r"\s*(<|>|,|::)\s*")
_classname_regularize_type = re.compile(
    r"[<,](" + "|".join([re.escape(p) for p in _root_alias_to_c_primitive]) + r")[>,]"
)

_classname_encode_pattern = re.compile(rb"[^a-zA-Z0-9]+")
_classname_decode_antiversion = re.compile(rb".*_([0-9a-f][0-9a-f])+_v([0-9]+)$")
_classname_decode_version = re.compile(rb".*_v([0-9]+)$")
_classname_decode_pattern = re.compile(rb"_(([0-9a-f][0-9a-f])+)_")


def _classname_decode_convert(hex_characters):
    g = hex_characters.group(1)
    return bytes(int(g[i : i + 2], 16) for i in range(0, len(g), 2))


def _classname_encode_convert(bad_characters):
    g = bad_characters.group(0)
    return b"_" + b"".join(f"{x:02x}".encode() for x in g) + b"_"


def classname_regularize(classname):
    """
    Removes spaces around ``<``, ``>``, and ``::`` characters in a classname
    so that they can be matched by string name.

    If ``classname`` is None, this function returns None. Otherwise, it must be
    a string and it returns a string.
    """
    if classname is not None:
        classname = re.sub(_classname_regularize, r"\1", classname)

        m = _classname_regularize_type.search(classname)

        while m is not None:
            start, stop = m.span(1)
            token = classname[start:stop]
            replacement = _root_alias_to_c_primitive[token]
            classname = classname[:start] + replacement + classname[stop:]

            m = _classname_regularize_type.search(classname)

    return classname


def classname_decode(encoded_classname):
    """
    Converts a Python (encoded) classname, such as ``Model_Some_3a3a_Thing``
    into a C++ (decoded) classname, such as ``Some::Thing``.

    C++ classnames can include namespace delimiters (``::``) and template
    arguments (``<`` and ``>``), which have to be translated into
    ``[A-Za-z_][A-Za-z0-9_]*`` for Python. Non-conforming characters and also
    underscores are translated to their hexadecimal equivalents and surrounded
    by underscores. Additionally, Python models of C++ classes are prepended
    with ``Model_`` (or ``Unknown_`` if a streamer isn't found).
    """
    if encoded_classname.startswith("Unknown_"):
        raw = encoded_classname[8:].encode()
    elif encoded_classname.startswith("Model_"):
        raw = encoded_classname[6:].encode()
    else:
        raise ValueError(f"not an encoded classname: {encoded_classname}")

    if _classname_decode_antiversion.match(raw) is not None:
        version = None
    else:
        m = _classname_decode_version.match(raw)
        if m is None:
            version = None
        else:
            version = int(m.group(1))
            raw = raw[: -len(m.group(1)) - 2]

    out = _classname_decode_pattern.sub(_classname_decode_convert, raw)
    return out.decode(), version


def classname_encode(classname, version=None, unknown=False):
    """
    Converts a C++ (decoded) classname, such as ``Some::Thing`` into a Python
    classname (encoded), such as ``Model_Some_3a3a_Thing``.

    If ``version`` is a number such as ``2``, the Python name is suffixed by
    version, such as ``Model_Some_3a3a_Thing_v2``.

    If ``unknown`` is True, the ``Model_`` prefix becomes ``Unknown_``.

    C++ classnames can include namespace delimiters (``::``) and template
    arguments (``<`` and ``>``), which have to be translated into
    ``[A-Za-z_][A-Za-z0-9_]*`` for Python. Non-conforming characters and also
    underscores are translated to their hexadecimal equivalents and surrounded
    by underscores. Additionally, Python models of C++ classes are prepended
    with ``Model_`` (or ``Unknown_`` if a streamer isn't found).
    """
    if unknown:
        prefix = "Unknown_"
    else:
        prefix = "Model_"
    if classname.startswith(prefix):
        raise ValueError(f"classname is already encoded: {classname}")

    if version is None:
        v = ""
    else:
        v = "_v" + str(version)

    raw = classname.encode()
    out = _classname_encode_pattern.sub(_classname_encode_convert, raw)
    return prefix + out.decode() + v


def classname_version(encoded_classname):
    """
    Extracts a version number from a Python (encoded) classname, if it has one.

    For example, ``Model_Some_3a3a_Thing_v2`` returns ``2``.

    A name without a version number, such as ``Model_Some_3a3a_Thing``, returns
    None.
    """
    raw = encoded_classname.encode()
    if _classname_decode_antiversion.match(raw) is not None:
        return None
    else:
        m = _classname_decode_version.match(raw)
        if m is None:
            return None
        else:
            return int(m.group(1))


def class_named(classname, version=None, custom_classes=None):
    """
    Returns a class with a given C++ (decoded) classname.

    If ``version`` is None, no attempt is made to find a specific version.

    * If the class is a :doc:`uproot.model.DispatchByVersion`, then this is
      object returned.
    * If the class is a versionless model, then this is the object returned.

    If ``version`` is an integer, an attempt is made to find the specific
    version.

    * If the class is a :doc:`uproot.model.DispatchByVersion`, then it is
      queried for a versioned model.
    * If the class is a versionless model, then this is the object returned.

    If ``custom_classes`` are provided, then these are searched (exclusively)
    for the class. If ``custom_classes`` is None, then ``uproot.classes`` is
    used.

    No classes are created if a class is not found (an error is raised).
    """
    if custom_classes is None:
        classes = uproot.classes
        where = "the 'custom_classes' dict"
    else:
        where = "uproot.classes"

    cls = classes.get(classname)
    if cls is None:
        raise ValueError(f"no class named {classname} in {where}")

    if version is not None and isinstance(cls, DispatchByVersion):
        versioned_cls = cls.class_of_version(version)
        if versioned_cls is not None:
            return versioned_cls
        else:
            raise ValueError(
                "no class named {} with version {} in {}".format(
                    classname, version, where
                )
            )

    else:
        return cls


def has_class_named(classname, version=None, custom_classes=None):
    """
    Returns True if :doc:`uproot.model.class_named` would find a class,
    False if it would raise an exception.
    """
    cls = maybe_custom_classes(classname, custom_classes).get(classname)
    if cls is None:
        return False

    if version is not None and isinstance(cls, DispatchByVersion):
        return cls.has_version(version)
    else:
        return True


def maybe_custom_classes(classname, custom_classes):
    """
    Passes through ``custom_classes`` if it is not None; returns
    ``uproot.classes`` otherwise.

    Some ``classnames`` are never custom (see ``uproot.model.never_from_streamers``).
    """
    if custom_classes is None or classname in never_from_streamers:
        return uproot.classes
    else:
        return custom_classes


class Model:
    """
    Abstract class for all objects extracted from ROOT files (except for
    :doc:`uproot.reading.ReadOnlyFile`, :doc:`uproot.reading.ReadOnlyDirectory`,
    and :doc:`uproot.reading.ReadOnlyKey`).

    A model is instantiated from a file using the :ref:`uproot.model.Model.read`
    classmethod or synthetically using the :ref:`uproot.model.Model.empty`
    classmethod, not through a normal constructor.

    Models point back to the file from which they were created, though only a
    few classes (named in ``uproot.reading.must_be_attached``) have an open,
    readable file attached; the rest have a :doc:`uproot.reading.DetachedFile`
    with information about the file, while not holding the file open.

    Uproot recognizes *some* of ROOT's thousands of classes, by way of methods
    and properties defined in :doc:`uproot.behaviors`. Examples include

    * :doc:`uproot.behaviors.TTree.TTree`
    * :doc:`uproot.behaviors.TH1.TH1`

    These classes are the most convenient to work with and have specialized
    documentation.

    Classes that don't have any predefined behaviors are still usable through
    their member data.

    * :ref:`uproot.model.Model.members`: a dict of C++ member names and values
      directly in this class.
    * :ref:`uproot.model.Model.all_members`: a dict of C++ member names and
      values in this class or any superclasses.
    * :ref:`uproot.model.Model.member`: method that takes a C++ member name
      and returns its value (from this or any superclass).
    * :ref:`uproot.model.Model.has_member`: method that takes a C++ member
      name and returns True if it exists (in this or any superclass), False
      otherwise.

    Accessing a data structure through its C++ members may be a prelude to
    adding custom behaviors for it. Before we know what conveniences to add, we
    need to know how they'll be used: this information comes from the user
    community.

    Pythonic models don't follow the same class inheritance tree as their C++
    counterparts: most of them are direct subclasses of
    :doc:`uproot.model.Model`, :doc:`uproot.model.DispatchByVersion`, or
    :doc:`uproot.model.VersionedModel`. To separate an object's members
    from its superclass members, a model instance is created for each and
    the superclass parts are included in a list called
    :ref:`uproot.model.Model.bases`.
    """

    base_names_versions = []
    member_names = []
    class_flags = {}
    class_code = None
    class_streamer = None
    class_rawstreamers = ()
    writable = False
    _deeply_writable = False
    behaviors = ()

    def __repr__(self):
        if self.class_version is None:
            version = ""
        else:
            version = f" (version {self.class_version})"
        return f"<{self.classname}{version} at 0x{id(self):012x}>"

    def __enter__(self):
        if isinstance(self._file, uproot.reading.ReadOnlyFile):
            self._file.source.__enter__()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if isinstance(self._file, uproot.reading.ReadOnlyFile):
            self._file.source.__exit__(exception_type, exception_value, traceback)

    @property
    def classname(self):
        """
        The C++ (decoded) classname of the modeled class.

        See :doc:`uproot.model.classname_decode`,
        :doc:`uproot.model.classname_encode`, and
        :doc:`uproot.model.classname_version`.
        """
        return classname_decode(self.encoded_classname)[0]

    @property
    def encoded_classname(self):
        """
        The Python (encoded) classname of the modeled class. May or may not
        include version.

        See :doc:`uproot.model.classname_decode`,
        :doc:`uproot.model.classname_encode`, and
        :doc:`uproot.model.classname_version`.
        """
        return type(self).__name__

    @property
    def class_version(self):
        """
        The version number of the modeled class (int) if any; None otherwise.

        See :doc:`uproot.model.classname_decode`,
        :doc:`uproot.model.classname_encode`, and
        :doc:`uproot.model.classname_version`.
        """
        return classname_decode(self.encoded_classname)[1]

    @property
    def cursor(self):
        """
        A cursor pointing to the start of this instance in the byte stream
        (before :ref:`uproot.model.Model.read_numbytes_version`).
        """
        return self._cursor

    @property
    def file(self):
        """
        A :doc:`uproot.reading.ReadOnlyFile`, which may be open and readable,
        or a :doc:`uproot.reading.DetachedFile`, which only contains
        information about the original file (not an open file handle).
        """
        return self._file

    def close(self):
        """
        Closes the file from which this object is derived, if such a file is
        still attached (i.e. not :doc:`uproot.reading.DetachedFile`).
        """
        if isinstance(self._file, uproot.reading.ReadOnlyFile):
            self._file.close()

    @property
    def closed(self):
        """
        True if the associated file is known to be closed; False if it is known
        to be open. If the associated file is detached
        (:doc:`uproot.reading.DetachedFile`), then the value is None.
        """
        if isinstance(self._file, uproot.reading.ReadOnlyFile):
            return self._file.closed
        else:
            return None

    @property
    def parent(self):
        """
        The object that was deserialized before this one in recursive descent,
        usually the containing object (or the container's container).
        """
        return self._parent

    @property
    def concrete(self):
        """
        The Python instance corresponding to the concrete (instantiated) class
        in C++, which is ``self`` if this is the concrete class or another
        object if this is actually a holder of superclass members for that other
        object (i.e. if this object is in the other's
        :ref:`uproot.model.Model.bases`).
        """
        if self._concrete is None:
            return self
        return self._concrete

    @property
    def members(self):
        """
        A dict of C++ member data directly associated with this class (i.e. not
        its superclasses). For all members, see
        :ref:`uproot.model.Model.all_members`.
        """
        return self._members

    @property
    def all_members(self):
        """
        A dict of C++ member data for this class and its superclasses. For only
        direct members, see :ref:`uproot.model.Model.members`.
        """
        out = {}
        for base in self._bases:
            out.update(base.all_members)
        out.update(self._members)
        return out

    def has_member(self, name, all=True):
        """
        Returns True if calling :ref:`uproot.model.Model.member` with the same
        arguments would return a value; False if the member is missing.
        """
        if name in self._members:
            return True
        if all:
            for base in reversed(self._bases):
                if base.has_member(name, all=True):
                    return True
        return False

    def member(self, name, all=True, none_if_missing=False):
        """
        Args:
            name (str): The name of the member datum to retrieve.
            all (bool): If True, recursively search all superclasses in
                :ref:`uproot.model.Model.bases`. Otherwise, search the
                direct class only.
            none_if_missing (bool): If a member datum doesn't exist in the
                search path, ``none_if_missing=True`` has this function return
                None, but ``none_if_missing=False`` would have it raise an
                exception. Note that None is a possible value for some member
                data.

        Returns a C++ member datum by name.
        """
        if name in self._members:
            return self._members[name]
        if all:
            for base in reversed(self._bases):
                if base.has_member(name, all=True):
                    return base.member(name, all=True)

        if none_if_missing:
            return None
        else:
            raise uproot.KeyInFileError(
                name,
                because="""{}.{} has only the following members:

    {}
""".format(
                    type(self).__module__,
                    type(self).__name__,
                    ", ".join(repr(x) for x in self.all_members),
                ),
                file_path=getattr(self._file, "file_path", None),
            )

    @property
    def bases(self):
        """
        List of :doc:`uproot.model.Model` objects representing superclass data
        for this object in the order given in C++ (opposite method resolution
        order).

        * If this object has no superclasses, ``bases`` is empty.
        * If it has one superclass, which itself might have superclasses,
          ``bases`` has length 1.
        * Only if this object *multiply inherits* from more than one superclass
          at the same level does ``bases`` have length greater than 1.

        Since multiple inheritance is usually avoided, ``bases`` rarely has
        length greater than 1. A linear chain of superclasses deriving from
        super-superclasses is represented by ``bases`` containing an object
        whose ``bases`` contains objects.
        """
        return self._bases

    def base(self, *cls):
        """
        Extracts instances from :ref:`uproot.model.Model.bases` by Python class
        type.

        The ``cls`` arguments may be Python classes or C++ classname strings to match.
        """
        cpp_names = [classname_regularize(x) for x in cls if uproot._util.isstr(x)]
        py_types = tuple(x for x in cls if not uproot._util.isstr(x))

        out = []
        for x in getattr(self, "_bases", []):
            if isinstance(x, py_types) or any(
                getattr(x, "classname", None) == n for n in cpp_names
            ):
                out.append(x)
            if isinstance(x, Model):
                out.extend(x.base(*cls))
        return out

    def is_instance(self, *cls):
        """
        Returns True if this object matches a given type in the C++ class hierarchy.

        The ``cls`` arguments may be Python classes or C++ classname strings to match.
        """
        cpp_names = [classname_regularize(x) for x in cls if uproot._util.isstr(x)]
        py_types = tuple(x for x in cls if not uproot._util.isstr(x))

        if isinstance(self, py_types) or any(self.classname == n for n in cpp_names):
            return True
        else:
            return len(self.base(*cls)) != 0

    @property
    def num_bytes(self):
        """
        Number of bytes expected in the (uncompressed) serialization of this
        instance.

        This value may be None (unknown before reading) or an integer.

        If the value is an integer and the object exists (no exceptions in
        :ref:`uproot.model.Model.read`), then the expected number of bytes
        agreed with the actual number of bytes, and this numer is reliable.

        If this object is re-serialized, it won't necessarily occupy the same
        number of bytes.
        """
        return self._num_bytes

    @property
    def instance_version(self):
        """
        Version of this instance as read from the byte stream.

        If this model is versioned (:doc:`uproot.model.VersionedModel`), the
        ``instance_version`` ought to be equal to the
        :ref:`uproot.model.Model.class_version`.

        If this model is versionless, the ``instance_version`` contains new
        information about the actual version deserialized.
        """
        return self._instance_version

    @property
    def is_memberwise(self):
        """
        True if the object was serialized in ROOT's memberwise format; False
        otherwise.
        """
        return self._is_memberwise

    @classmethod
    def awkward_form(cls, file, context):
        """
        Args:
            cls (subclass of :doc:`uproot.model.Model`): This class.
            file (:doc:`uproot.reading.ReadOnlyFile`): File to use to generate
                :doc:`uproot.model.Model` classes from its
                :ref:`uproot.reading.ReadOnlyFile.streamers` and ``file_path``
                for error messages.
            context (dict): Context for the Form-generation; defaults are
                the remaining arguments below.
            index_format (str): Format to use for indexes of the
                ``awkward.forms.Form``; may be ``"i32"``, ``"u32"``, or
                ``"i64"``.
            header (bool): If True, include header fields of each C++ class.
            tobject_header (bool): If True, include header fields of each ``TObject``
                base class.
            breadcrumbs (tuple of class objects): Used to check for recursion.
                Types that contain themselves cannot be Awkward Arrays because the
                depth of instances is unknown.

        The ``awkward.forms.Form`` to use to put objects of type type in an
        Awkward Array.
        """
        raise uproot.interpretation.objects.CannotBeAwkward(
            classname_decode(cls.__name__)[0]
        )

    @classmethod
    def strided_interpretation(
        cls, file, header=False, tobject_header=True, breadcrumbs=(), original=None
    ):
        """
        Args:
            cls (subclass of :doc:`uproot.model.Model`): This class.
            file (:doc:`uproot.reading.ReadOnlyFile`): File to use to generate
                :doc:`uproot.model.Model` classes from its
                :ref:`uproot.reading.ReadOnlyFile.streamers` and ``file_path``
                for error messages.
            header (bool): If True, assume the outermost object has a header.
            tobject_header (bool): If True, assume that ``TObjects`` have headers.
            original (None, :doc:`uproot.model.Model`, or :doc:`uproot.containers.Container`): The
                original, non-strided model or container.
            breadcrumbs (tuple of class objects): Used to check for recursion.
                Types that contain themselves cannot be strided because the
                depth of instances is unknown.

        Returns a list of (str, ``numpy.dtype``) pairs to build a
        :doc:`uproot.interpretation.objects.AsStridedObjects` interpretation.
        """
        raise uproot.interpretation.objects.CannotBeStrided(
            classname_decode(cls.__name__)[0]
        )

    def tojson(self):
        """
        Serializes this object in its ROOT JSON form (as Python lists and dicts,
        which can be passed to ``json.dump`` or ``json.dumps``).
        """
        out = {}
        for base in self._bases:
            tmp = base.tojson()
            if isinstance(tmp, dict):
                out.update(tmp)
        for k, v in self.members.items():
            if isinstance(v, Model):
                out[k] = v.tojson()
            elif isinstance(v, (numpy.number, numpy.ndarray)):
                out[k] = v.tolist()
            else:
                out[k] = v
        out["_typename"] = self.classname
        return out

    @classmethod
    def empty(cls):
        """
        Creates a model instance (of subclass ``cls``) with no data; all
        required attributes are None or empty.
        """
        self = cls.__new__(cls)
        self._cursor = None
        self._file = None
        self._parent = None
        self._concrete = None
        self._members = {}
        self._bases = []
        self._num_bytes = None
        self._instance_version = None
        self._is_memberwise = False
        return self

    @classmethod
    def read(cls, chunk, cursor, context, file, selffile, parent, concrete=None):
        """
        Args:
            cls (subclass of :doc:`uproot.model.Model`): Class to instantiate.
            chunk (:doc:`uproot.source.chunk.Chunk`): Buffer of contiguous data
                from the file :doc:`uproot.source.chunk.Source`.
            cursor (:doc:`uproot.source.cursor.Cursor`): Current position in
                that ``chunk``.
            context (dict): Auxiliary data used in deserialization.
            file (:doc:`uproot.reading.ReadOnlyFile`): An open file object,
                capable of generating new :doc:`uproot.model.Model` classes
                from its :ref:`uproot.reading.ReadOnlyFile.streamers`.
            selffile (:doc:`uproot.reading.CommonFileMethods`): A possibly
                :doc:`uproot.reading.DetachedFile` associated with this object.
            parent (None or calling object): The previous ``read`` in the
                recursive descent.
            concrete (None or :doc:`uproot.model.Model` instance): If None,
                this model corresponds to the concrete (instantiated) class in
                C++. Otherwise, this model represents a superclass part of the
                object, and ``concrete`` points to the concrete instance.

        Creates a model instance by reading data from a file.
        """
        self = cls.__new__(cls)
        self._cursor = cursor.copy()
        self._file = selffile
        self._parent = parent
        self._concrete = concrete
        self._members = {}
        self._bases = []
        self._num_bytes = None
        self._instance_version = None
        self._is_memberwise = False
        old_breadcrumbs = context.get("breadcrumbs", ())
        context["breadcrumbs"] = old_breadcrumbs + (self,)

        self.hook_before_read(chunk=chunk, cursor=cursor, context=context, file=file)
        forth_stash = uproot._awkward_forth.forth_stash(context)
        if forth_stash is not None:
            forth_obj = context["forth"].gen

        if context.get("reading", True):
            temp_index = cursor._index
            self.read_numbytes_version(chunk, cursor, context)
            length = cursor._index - temp_index
            if length != 0:
                if forth_stash is not None:
                    forth_stash.add_to_pre(f"{length} stream skip\n")
            if (
                issubclass(cls, VersionedModel)
                and self._instance_version != classname_version(cls.__name__)
                and self._instance_version is not None
            ):
                correct_cls = file.class_named(self.classname, self._instance_version)
                if classname_version(correct_cls.__name__) != classname_version(
                    cls.__name__
                ):
                    if forth_stash is not None:
                        forth_obj.add_node(
                            "pass",
                            forth_stash.get_attrs(),
                            "i64",
                            1,
                            {},
                        )
                    cursor.move_to(self._cursor.index)
                    context["breadcrumbs"] = old_breadcrumbs
                    temp_var = correct_cls.read(
                        chunk,
                        cursor,
                        context,
                        file,
                        selffile,
                        parent,
                        concrete=concrete,
                    )
                    # if forth_stash is not None:
                    #    forth_obj.go_to(temp)
                    return temp_var

        if context.get("in_TBranch", False):
            # AwkwardForth testing: test_0637's 01,02,05,08,09,11,12,13,15,16,29,35,38,39,44,45,46,47,49,50,52,56
            if self._num_bytes is None and self._instance_version != self.class_version:
                self._instance_version = None
                cursor = self._cursor
                if forth_stash is not None and not context["cancel_forth"]:
                    forth_stash._pre_code.pop(-1)

            elif self._instance_version == 0:
                if forth_stash is not None:
                    forth_stash.add_to_pre("4 stream skip\n")
                cursor.skip(4)

        if context.get("reading", True):
            self.hook_before_read_members(
                chunk=chunk, cursor=cursor, context=context, file=file
            )
            if forth_stash is not None:
                forth_obj.add_node(
                    "model828",
                    forth_stash.get_attrs(),
                    "i64",
                    1,
                    {},
                )

            self.read_members(chunk, cursor, context, file)
            self.hook_after_read_members(
                chunk=chunk, cursor=cursor, context=context, file=file
            )

        self.check_numbytes(chunk, cursor, context)

        self.hook_before_postprocess(
            chunk=chunk, cursor=cursor, context=context, file=file
        )

        out = self.postprocess(chunk, cursor, context, file)

        context["breadcrumbs"] = old_breadcrumbs

        return out

    def read_numbytes_version(self, chunk, cursor, context):
        """
        Args:
            chunk (:doc:`uproot.source.chunk.Chunk`): Buffer of contiguous data
                from the file :doc:`uproot.source.chunk.Source`.
            cursor (:doc:`uproot.source.cursor.Cursor`): Current position in
                that ``chunk``.
            context (dict): Auxiliary data used in deserialization.

        Reads the number of bytes and instance version from the byte stream,
        which is usually 6 bytes (4 + 2). Bits with special meanings are
        appropriately masked out.

        Some types don't have a 6-byte header or handle it differently; in
        those cases, this method should be overridden.
        """
        import uproot.deserialization

        (
            self._num_bytes,
            self._instance_version,
            self._is_memberwise,
        ) = uproot.deserialization.numbytes_version(chunk, cursor, context)

    def read_members(self, chunk, cursor, context, file):
        """
        Args:
            chunk (:doc:`uproot.source.chunk.Chunk`): Buffer of contiguous data
                from the file :doc:`uproot.source.chunk.Source`.
            cursor (:doc:`uproot.source.cursor.Cursor`): Current position in
                that ``chunk``.
            context (dict): Auxiliary data used in deserialization.
            file (:doc:`uproot.reading.ReadOnlyFile`): An open file object,
                capable of generating new :doc:`uproot.model.Model` classes
                from its :ref:`uproot.reading.ReadOnlyFile.streamers`.

        Reads the member data for this class. The abstract class
        :doc:`uproot.model.Model` has an empty ``read_members`` method; this
        *must* be overridden by subclasses.
        """
        pass

    def check_numbytes(self, chunk, cursor, context):
        """
        Args:
            chunk (:doc:`uproot.source.chunk.Chunk`): Buffer of contiguous data
                from the file :doc:`uproot.source.chunk.Source`.
            cursor (:doc:`uproot.source.cursor.Cursor`): Current position in
                that ``chunk``.
            context (dict): Auxiliary data used in deserialization.

        Reads nothing; checks the expected number of bytes against the actual
        movement of the ``cursor`` at the end of the object, possibly raising
        a :doc:`uproot.deserialization.DeserializationError` exception.

        If :ref:`uproot.model.Model.num_bytes` is None, this method does
        nothing.

        It is *possible* that a subclass would override this method, but not
        likely.
        """
        import uproot.deserialization

        uproot.deserialization.numbytes_check(
            chunk,
            self._cursor,
            cursor,
            self._num_bytes,
            self.classname,
            context,
            getattr(self._file, "file_path", None),
        )

    def postprocess(self, chunk, cursor, context, file):
        """
        Args:
            chunk (:doc:`uproot.source.chunk.Chunk`): Buffer of contiguous data
                from the file :doc:`uproot.source.chunk.Source`.
            cursor (:doc:`uproot.source.cursor.Cursor`): Current position in
                that ``chunk``.
            context (dict): Auxiliary data used in deserialization.
            file (:doc:`uproot.reading.ReadOnlyFile`): An open file object,
                capable of generating new :doc:`uproot.model.Model` classes
                from its :ref:`uproot.reading.ReadOnlyFile.streamers`.

        Called for any additional processing after the object has been fully
        read.

        The return value from this method is the object that actually represents
        the ROOT data, which might be a different instance or even a different
        type from this class. The default in :doc:`uproot.model.Model` is to
        return ``self``.

        Note that for versioned models,
        :ref:`uproot.model.VersionedModel.postprocess` is called first, then
        :ref:`uproot.model.DispatchByVersion.postprocess` is called on its
        output, allowing a :doc:`uproot.model.DispatchByVersion` to refine all
        data of its type, regardless of version.
        """
        return self

    def hook_before_read(self, **kwargs):
        """
        Called in :ref:`uproot.model.Model.read`, before any data have been
        read.
        """
        pass

    def hook_before_read_members(self, **kwargs):
        """
        Called in :ref:`uproot.model.Model.read`, after
        :ref:`uproot.model.Model.read_numbytes_version` and before
        :ref:`uproot.model.Model.read_members`.
        """
        pass

    def hook_after_read_members(self, **kwargs):
        """
        Called in :ref:`uproot.model.Model.read`, after
        :ref:`uproot.model.Model.read_members` and before
        :ref:`uproot.model.Model.check_numbytes`.
        """
        pass

    def hook_before_postprocess(self, **kwargs):
        """
        Called in :ref:`uproot.model.Model.read`, after
        :ref:`uproot.model.Model.check_numbytes` and before
        :ref:`uproot.model.Model.postprocess`.
        """
        pass

    def _to_writable_postprocess(self, original):
        pass

    def _to_writable(self, concrete):
        cls = None
        if self.writable:
            cls = type(self)
        else:
            unversioned = uproot.classes.get(self.classname)
            if issubclass(unversioned, DispatchByVersion):
                for versioned_cls in unversioned.known_versions.values():
                    if versioned_cls.writable:
                        cls = versioned_cls
                        break

            elif unversioned is not None:
                if unversioned.writable:
                    cls = unversioned

        if cls is None:
            raise NotImplementedError(
                f"this ROOT type is not writable: {self.classname}"
            )
        else:
            out = cls.__new__(cls)
            out._cursor = self._cursor
            out._file = self._file
            out._parent = self._parent
            out._concrete = concrete
            out._num_bytes = self._num_bytes
            out._instance_version = classname_decode(cls.__name__)[1]
            out._is_memberwise = self._is_memberwise

            if concrete is None:
                concrete = out

            out._bases = []
            for base in self._bases:
                out._bases.append(base._to_writable(concrete))

            out._members = {}
            for key, value in self._members.items():
                if isinstance(value, Model):
                    out._members[key] = value._to_writable(None)
                else:
                    out._members[key] = value

            out._to_writable_postprocess(self)
            return out

    def to_writable(self):
        """
        Args:
            obj (:doc:`uproot.model.Model` instance of the same C++ class): The
                object to convert to this class version.

        Returns a writable version of this object or raises a NotImplementedError
        if no writable version exists.
        """
        if self._deeply_writable:
            return self
        else:
            return self._to_writable(None)

    def _serialize(self, out, header, name, tobject_flags):
        raise NotImplementedError(
            "can't write {} instances yet ('serialize' method not implemented)".format(
                type(self).__name__
            )
        )

    def serialize(self, name=None):
        """
        Serialize a object (from num_bytes and version onward) for writing into
        an output ROOT file.

        If a ``name`` is given, override the object's current name.

        This method has not been implemented on all classes (raises
        NotImplementedError).
        """
        out = []
        self._serialize(out, True, name, numpy.uint32(0x00000000))
        return b"".join(out)

    def to_pyroot(self, name=None):
        """
        Args:
            name (str or None): A name for the new PyROOT object.

        Converts this :doc:`uproot.model.Model` into a PyROOT object *if it is writable*.
        A minority of Uproot models are writable, mostly just histograms. Writability
        is necessary for conversion to PyROOT because it is serialized through a
        ROOT TMessage.
        """
        return uproot.pyroot.to_pyroot(self, name=name)


class VersionedModel(Model):
    """
    A Python class that models a specific version of a ROOT C++ class.

    Classes that inherit directly from :doc:`uproot.model.Model` are versionless,
    classes that inherit from :doc:`uproot.model.VersionedModel` depend on
    version.

    Note that automatically generated :doc:`uproot.model.VersionedModel` classes
    are placed in the ``uproot.dynamic`` namespace. This namespace can generate
    :doc:`uproot.model.DynamicModel` classes on demand in Python 3.7 and above,
    which automatically generated :doc:`uproot.model.VersionedModel` classes
    rely upon to be pickleable. Therefore, ROOT object types without predefined
    :doc:`uproot.model.Model` classes cannot be pickled in Python versions
    before 3.7.
    """

    def __getstate__(self):
        return (
            {
                "base_names_versions": self.base_names_versions,
                "member_names": self.member_names,
                "class_flags": self.class_flags,
                "class_code": self.class_code,
                "class_streamer": self.class_streamer,
                "class_rawstreamers": self.class_rawstreamers,
                "writable": self.writable,
                "behaviors": self.behaviors,
            },
            dict(self.__dict__),
        )

    def __setstate__(self, state):
        class_data, instance_data = state
        self.__dict__.update(instance_data)


class DispatchByVersion:
    """
    A Python class that models all versions of a ROOT C++ class by maintaining
    a dict of :doc:`uproot.model.VersionedModel` classes.

    The :ref:`uproot.model.DispatchByVersion.read` classmethod reads the
    instance version number from the byte stream, backs up the
    :doc:`uproot.source.cursor.Cursor` to the starting position, and invokes
    the appropriate :doc:`uproot.model.VersionedModel`'s ``read`` classmethod.

    If a :doc:`uproot.model.VersionedModel` does not exist for the specified
    version, the ``file``'s ``TStreamerInfo`` is queried to attempt to create
    one, and failing that, an :doc:`uproot.model.UnknownClassVersion` is
    created instead.

    Note that :doc:`uproot.model.DispatchByVersion` is not a subclass of
    :doc:`uproot.model.Model`. Instances of this class are not usable as
    stand-ins for ROOT data.
    """

    @classmethod
    def awkward_form(cls, file, context):
        """
        Args:
            cls (subclass of :doc:`uproot.model.DispatchByVersion`): This class.
            file (:doc:`uproot.reading.ReadOnlyFile`): File to use to generate
                :doc:`uproot.model.Model` classes from its
                :ref:`uproot.reading.ReadOnlyFile.streamers` and ``file_path``
                for error messages.
            context (dict): Context for the Form-generation; defaults are
                ``{"index_format": "i64", "header": False, "tobject_header": True, "breadcrumbs": ()}``.
                See below for context argument descriptions.
            index_format (str): Format to use for indexes of the
                ``awkward.forms.Form``; may be ``"i32"``, ``"u32"``, or
                ``"i64"``.
            header (bool): If True, include headers in the Form's ``"uproot"``
                parameters.
            tobject_header (bool): If True, include headers for ``TObject``
                classes in the Form's ``"uproot"`` parameters.
            breadcrumbs (tuple of class objects): Used to check for recursion.
                Types that contain themselves cannot be Awkward Arrays because the
                depth of instances is unknown.

        The ``awkward.forms.Form`` to use to put objects of type type in an
        Awkward Array.
        """
        versioned_cls = file.class_named(classname_decode(cls.__name__)[0], "max")
        return versioned_cls.awkward_form(file, context)

    @classmethod
    def strided_interpretation(
        cls, file, header=False, tobject_header=True, breadcrumbs=(), original=None
    ):
        """
        Args:
            cls (subclass of :doc:`uproot.model.DispatchByVersion`): This class.
            file (:doc:`uproot.reading.ReadOnlyFile`): File to use to generate
                :doc:`uproot.model.Model` classes from its
                :ref:`uproot.reading.ReadOnlyFile.streamers` and ``file_path``
                for error messages.
            header (bool): If True, assume the outermost object has a header.
            tobject_header (bool): If True, assume that ``TObjects`` have headers.
            original (None, :doc:`uproot.model.Model`, or :doc:`uproot.containers.Container`): The
                original, non-strided model or container.
            breadcrumbs (tuple of class objects): Used to check for recursion.
                Types that contain themselves cannot be strided because the
                depth of instances is unknown.

        Returns a list of (str, ``numpy.dtype``) pairs to build a
        :doc:`uproot.interpretation.objects.AsStridedObjects` interpretation.
        """
        versioned_cls = file.class_named(classname_decode(cls.__name__)[0], "max")
        return versioned_cls.strided_interpretation(
            file, header=header, tobject_header=tobject_header, breadcrumbs=breadcrumbs
        )

    @classmethod
    def class_of_version(cls, version):
        """
        Returns the class corresponding to a specified ``version`` if it exists.

        If not, this classmethod returns None. No attempt is made to create a
        missing class.
        """
        out = cls.known_versions.get(version)
        if out is None and version == 0 and len(cls.known_versions) != 0:
            out = cls.known_versions[max(cls.known_versions)]
        return out

    @classmethod
    def has_version(cls, version):
        """
        Returns True if a class corresponding to a specified ``version``
        currently exists; False otherwise.
        """
        return version in cls.known_versions

    @classmethod
    def new_class(cls, file, version):
        """
        Uses ``file`` to create a new class for a specified ``version``.

        As a side-effect, this new class is added to ``cls.known_versions``
        (for :ref:`uproot.model.DispatchByVersion.class_of_version` and
        :ref:`uproot.model.DispatchByVersion.has_version`).

        If the ``file`` lacks a ``TStreamerInfo`` for the class, this function
        returns a :doc:`uproot.model.UnknownClassVersion` (adding it to
        ``uproo4.unknown_classes`` if it's not already there).
        """
        classname, _ = classname_decode(cls.__name__)
        classname = classname_regularize(classname)
        streamer = file.streamer_named(classname, version)

        if streamer is None:
            streamer = file.streamer_named(classname, "max")

        if streamer is None and file.custom_classes is not None:
            cls2 = uproot.classes.get(classname)
            versioned_cls2 = cls2.class_of_version(version)
            if versioned_cls2 is not None:
                return versioned_cls2

        if streamer is not None:
            versioned_cls = streamer.new_class(file)
            versioned_cls.class_streamer = streamer
            cls.known_versions[streamer.class_version] = versioned_cls
            return versioned_cls

        else:
            unknown_cls = uproot.unknown_classes.get(classname)
            if unknown_cls is None:
                unknown_cls = uproot._util.new_class(
                    classname_encode(classname, version, unknown=True),
                    (UnknownClassVersion,),
                    {},
                )
                uproot.unknown_classes[classname] = unknown_cls
            return unknown_cls

    @classmethod
    def read(cls, chunk, cursor, context, file, selffile, parent, concrete=None):
        """
        Args:
            cls (subclass of :doc:`uproot.model.DispatchByVersion`): This class.
            chunk (:doc:`uproot.source.chunk.Chunk`): Buffer of contiguous data
                from the file :doc:`uproot.source.chunk.Source`.
            cursor (:doc:`uproot.source.cursor.Cursor`): Current position in
                that ``chunk``.
            context (dict): Auxiliary data used in deserialization.
            file (:doc:`uproot.reading.ReadOnlyFile`): An open file object,
                capable of generating new :doc:`uproot.model.Model` classes
                from its :ref:`uproot.reading.ReadOnlyFile.streamers`.
            selffile (:doc:`uproot.reading.CommonFileMethods`): A possibly
                :doc:`uproot.reading.DetachedFile` associated with this object.
            parent (None or calling object): The previous ``read`` in the
                recursive descent.
            concrete (None or :doc:`uproot.model.Model` instance): If None,
                this model corresponds to the concrete (instantiated) class in
                C++. Otherwise, this model represents a superclass part of the
                object, and ``concrete`` points to the concrete instance.

        Reads the instance version number from the byte stream, backs up the
        :doc:`uproot.source.cursor.Cursor` to the starting position, and
        invokes the appropriate :doc:`uproot.model.VersionedModel`'s ``read``
        classmethod.

        If a :doc:`uproot.model.VersionedModel` does not exist for the
        specified version, the ``file``'s ``TStreamerInfo`` is queried to
        attempt to create one, and failing that, an
        :doc:`uproot.model.UnknownClassVersion` is created instead.
        """
        import uproot.deserialization

        forth_stash = uproot._awkward_forth.forth_stash(context)

        if forth_stash is not None:
            forth_obj = forth_stash.get_gen_obj()
        # Ignores context["reading"], because otherwise, there would be nothing to do.
        start_index = cursor._index
        (
            num_bytes,
            version,
            is_memberwise,
        ) = uproot.deserialization.numbytes_version(chunk, cursor, context, move=False)

        versioned_cls = cls.class_of_version(version)
        bytes_skipped = cursor._index - start_index
        if forth_stash is not None:
            # raise NotImplementedError
            forth_stash.add_to_pre(f"{bytes_skipped} stream skip \n")
            forth_obj.add_node(
                "Model1319",
                forth_stash.get_attrs(),
                "i64",
                1,
                {},
            )

        if versioned_cls is not None:
            pass

        elif version is not None:
            versioned_cls = cls.new_class(file, version)

        elif context.get("in_TBranch", False):
            versioned_cls = cls.new_class(file, "max")

        else:
            raise ValueError(
                """Unknown version {} for class {} that cannot be skipped """
                """because its number of bytes is unknown.
""".format(
                    version,
                    classname_decode(cls.__name__)[0],
                )
            )

        # versioned_cls.read starts with numbytes_version again because move=False (above)
        # if forth_stash is not None:
        temp_var = cls.postprocess(
            versioned_cls.read(
                chunk, cursor, context, file, selffile, parent, concrete=concrete
            ),
            chunk,
            cursor,
            context,
            file,
        )
        # if forth_stash is not None:
        #    if "no_go_to" not in context.keys():
        # raise NotImplementedError
        # forth_obj.go_to(temp_node)
        return temp_var

    @classmethod
    def postprocess(cls, self, chunk, cursor, context, file):
        """
        Args:
            cls (subclass of :doc:`uproot.model.DispatchByVersion`): This class.
            chunk (:doc:`uproot.source.chunk.Chunk`): Buffer of contiguous data
                from the file :doc:`uproot.source.chunk.Source`.
            cursor (:doc:`uproot.source.cursor.Cursor`): Current position in
                that ``chunk``.
            context (dict): Auxiliary data used in deserialization.
            file (:doc:`uproot.reading.ReadOnlyFile`): An open file object,
                capable of generating new :doc:`uproot.model.Model` classes
                from its :ref:`uproot.reading.ReadOnlyFile.streamers`.

        Called for any additional processing after the object has been fully
        read.

        The return value from this method is the object that actually represents
        the ROOT data, which might be a different instance or even a different
        type from this class. The default in :doc:`uproot.model.Model` is to
        return ``self``.

        Note that for versioned models,
        :ref:`uproot.model.VersionedModel.postprocess` is called first, then
        :ref:`uproot.model.DispatchByVersion.postprocess` is called on its
        output, allowing a :doc:`uproot.model.DispatchByVersion` to refine all
        data of its type, regardless of version.
        """
        return self


class UnknownClass(Model):
    """
    Placeholder for a C++ class instance that has no
    :doc:`uproot.model.DispatchByVersion` and no ``TStreamerInfo`` in the
    current :doc:`uproot.reading.ReadOnlyFile` to produce one.
    """

    @property
    def chunk(self):
        """
        The ``chunk`` of data associated with the unknown class, referred to by
        a weak reference (to avoid memory leaks in
        :doc:`uproot.model.UnknownClass` objects). If the original ``chunk``
        has been garbage-collected, this raises ``RuntimeError``.

        Primarily useful in the :ref:`uproot.model.UnknownClass.debug` method.
        """
        chunk = self._chunk()
        if chunk is None:
            raise RuntimeError(
                "the 'chunk' associated with this unknown class has been deleted"
            )
        else:
            return chunk

    @property
    def context(self):
        """
        The auxiliary data used in deserialization.

        Primarily useful in the :ref:`uproot.model.UnknownClass.debug` method.
        """
        return self._context

    def __repr__(self):
        return f"<Unknown {self.classname} at 0x{id(self):012x}>"

    def debug(
        self, skip_bytes=0, limit_bytes=None, dtype=None, offset=0, stream=sys.stdout
    ):
        """
        Args:
            skip_bytes (int): Number of bytes to skip before presenting the
                remainder of the :doc:`uproot.source.chunk.Chunk`. May be
                negative, to examine the byte stream leading up to the attempted
                instantiation. The default, ``0``, starts where the number
                of bytes and version number would be (just before
                :ref:`uproot.model.Model.read_numbytes_version`).
            limit_bytes (None or int): Number of bytes to limit the output to.
                A line of debugging output (without any ``offset``) is 20 bytes,
                so multiples of 20 show full lines. If None, everything is
                shown to the end of the :doc:`uproot.source.chunk.Chunk`,
                which might be large.
            dtype (None, ``numpy.dtype``, or its constructor argument): If None,
                present only the bytes as decimal values (0-255). Otherwise,
                also interpret them as an array of a given NumPy type.
            offset (int): Number of bytes to skip before interpreting a ``dtype``;
                can be helpful if the numerical values are out of phase with
                the first byte shown. Not to be confused with ``skip_bytes``,
                which determines which bytes are shown at all. Any ``offset``
                values that are equivalent modulo ``dtype.itemsize`` show
                equivalent interpretations.
            stream (object with a ``write(str)`` method): Stream to write the
                debugging output to.

        Presents the byte stream at the point where this instance would have been
        deserialized.

        Example output with ``dtype=">f4"`` and ``offset=3``.

        .. code-block::

            --+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
            123 123 123  63 140 204 205  64  12 204 205  64  83  51  51  64 140 204 205  64
              {   {   {   ? --- --- ---   @ --- --- ---   @   S   3   3   @ --- --- ---   @
                                    1.1             2.2             3.3             4.4
                --+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
                176   0   0  64 211  51  51  64 246 102 102  65  12 204 205  65  30 102 102  66
                --- --- ---   @ ---   3   3   @ ---   f   f   A --- --- ---   A ---   f   f   B
                        5.5             6.6             7.7             8.8             9.9
                --+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
                202   0   0  67  74   0   0  67 151 128   0 123 123
                --- --- ---   C   J --- ---   C --- --- ---   {   {
                      101.0           202.0           303.0
        """
        cursor = self._cursor.copy()
        cursor.skip(skip_bytes)
        cursor.debug(
            self.chunk,
            context=self._context,
            limit_bytes=limit_bytes,
            dtype=dtype,
            offset=offset,
            stream=stream,
        )

    def debug_array(self, skip_bytes=0, dtype=np_uint8):
        """
        Args:
            skip_bytes (int): Number of bytes to skip before presenting the
                remainder of the :doc:`uproot.source.chunk.Chunk`. May be
                negative, to examine the byte stream leading up to the attempted
                instantiation. The default, ``0``, starts where the number
                of bytes and version number would be (just before
                :ref:`uproot.model.Model.read_numbytes_version`).
            dtype (``numpy.dtype`` or its constructor argument): Data type in
                which to interpret the data. (The size of the array returned is
                truncated to this ``dtype.itemsize``.)

        Like :ref:`uproot.model.UnknownClass.debug`, but returns a NumPy array
        for further inspection.
        """
        dtype = numpy.dtype(dtype)
        cursor = self._cursor.copy()
        cursor.skip(skip_bytes)
        out = self.chunk.remainder(cursor.index, cursor, self._context)
        return out[: (len(out) // dtype.itemsize) * dtype.itemsize].view(dtype)

    def read_members(self, chunk, cursor, context, file):
        self._chunk = weakref.ref(chunk)
        self._context = context

        if self._num_bytes is not None:
            cursor.skip(self._num_bytes - cursor.displacement(self._cursor))

        else:
            raise ValueError(
                """unknown class {} that cannot be skipped because its """
                """number of bytes is unknown
in file {}""".format(
                    self.classname, file.file_path
                )
            )


class UnknownClassVersion(VersionedModel):
    """
    Placeholder for a C++ class instance that has no ``TStreamerInfo`` in the
    current :doc:`uproot.reading.ReadOnlyFile` to produce one.
    """

    @property
    def chunk(self):
        """
        The ``chunk`` of data associated with the class of unknown version,
        referred to by a weak reference (to avoid memory leaks in
        :doc:`uproot.model.UnknownClassVersion` objects). If the original
        ``chunk`` has been garbage-collected, this raises ``RuntimeError``.

        Primarily useful in the :ref:`uproot.model.UnknownClassVersion.debug`
        method.
        """
        chunk = self._chunk()
        if chunk is None:
            raise RuntimeError(
                "the 'chunk' associated with this class of unknown version has "
                "been deleted"
            )
        else:
            return chunk

    @property
    def context(self):
        """
        The auxiliary data used in deserialization.

        Primarily useful in the :ref:`uproot.model.UnknownClass.debug` method.
        """
        return self._context

    def debug(
        self, skip_bytes=0, limit_bytes=None, dtype=None, offset=0, stream=sys.stdout
    ):
        """
        Args:
            skip_bytes (int): Number of bytes to skip before presenting the
                remainder of the :doc:`uproot.source.chunk.Chunk`. May be
                negative, to examine the byte stream leading up to the attempted
                instantiation. The default, ``0``, starts where the number
                of bytes and version number would be (just before
                :ref:`uproot.model.Model.read_numbytes_version`).
            limit_bytes (None or int): Number of bytes to limit the output to.
                A line of debugging output (without any ``offset``) is 20 bytes,
                so multiples of 20 show full lines. If None, everything is
                shown to the end of the :doc:`uproot.source.chunk.Chunk`,
                which might be large.
            dtype (None, ``numpy.dtype``, or its constructor argument): If None,
                present only the bytes as decimal values (0-255). Otherwise,
                also interpret them as an array of a given NumPy type.
            offset (int): Number of bytes to skip before interpreting a ``dtype``;
                can be helpful if the numerical values are out of phase with
                the first byte shown. Not to be confused with ``skip_bytes``,
                which determines which bytes are shown at all. Any ``offset``
                values that are equivalent modulo ``dtype.itemsize`` show
                equivalent interpretations.
            stream (object with a ``write(str)`` method): Stream to write the
                debugging output to.

        Presents the byte stream at the point where this instance would have been
        deserialized.

        Example output with ``dtype=">f4"`` and ``offset=3``.

        .. code-block::

            --+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
            123 123 123  63 140 204 205  64  12 204 205  64  83  51  51  64 140 204 205  64
              {   {   {   ? --- --- ---   @ --- --- ---   @   S   3   3   @ --- --- ---   @
                                    1.1             2.2             3.3             4.4
                --+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
                176   0   0  64 211  51  51  64 246 102 102  65  12 204 205  65  30 102 102  66
                --- --- ---   @ ---   3   3   @ ---   f   f   A --- --- ---   A ---   f   f   B
                        5.5             6.6             7.7             8.8             9.9
                --+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
                202   0   0  67  74   0   0  67 151 128   0 123 123
                --- --- ---   C   J --- ---   C --- --- ---   {   {
                      101.0           202.0           303.0
        """
        cursor = self._cursor.copy()
        cursor.skip(skip_bytes)
        cursor.debug(
            self.chunk,
            context=self._context,
            limit_bytes=limit_bytes,
            dtype=dtype,
            offset=offset,
            stream=stream,
        )

    def debug_array(self, skip_bytes=0, dtype=np_uint8):
        """
        Args:
            skip_bytes (int): Number of bytes to skip before presenting the
                remainder of the :doc:`uproot.source.chunk.Chunk`. May be
                negative, to examine the byte stream leading up to the attempted
                instantiation. The default, ``0``, starts where the number
                of bytes and version number would be (just before
                :ref:`uproot.model.Model.read_numbytes_version`).
            dtype (``numpy.dtype`` or its constructor argument): Data type in
                which to interpret the data. (The size of the array returned is
                truncated to this ``dtype.itemsize``.)

        Like :ref:`uproot.model.UnknownClassVersion.debug`, but returns a
        NumPy array for further inspection.
        """
        dtype = numpy.dtype(dtype)
        cursor = self._cursor.copy()
        cursor.skip(skip_bytes)
        out = self.chunk.remainder(cursor.index, cursor, self._context)
        return out[: (len(out) // dtype.itemsize) * dtype.itemsize].view(dtype)

    def read_members(self, chunk, cursor, context, file):
        self._chunk = weakref.ref(chunk)
        self._context = context

        if self._num_bytes is not None:
            cursor.skip(self._num_bytes - cursor.displacement(self._cursor))

        else:
            raise ValueError(
                """class {} with unknown version {} cannot be skipped """
                """because its number of bytes is unknown
in file {}""".format(
                    self.classname, self._instance_version, file.file_path
                )
            )

    def __repr__(self):
        return "<{} with unknown version {} at 0x{:012x}>".format(
            self.classname, self._instance_version, id(self)
        )


class DynamicModel(VersionedModel):
    """
    A :doc:`uproot.model.VersionedModel` subclass generated by any attempt to
    extract it from the ``uproot.dynamic`` namespace in Python 3.7 and later.

    This dynamically generated model allows ROOT object types without predefined
    :doc:`uproot.model.Model` classes to be pickled in Python 3.7 and later.
    """

    def __setstate__(self, state):
        cls = type(self)
        class_data, instance_data = state
        for k, v in class_data.items():
            if not hasattr(cls, k):
                setattr(cls, k, v)
        cls.__bases__ = (
            tuple(x for x in class_data["behaviors"] if x not in cls.__bases__)
            + cls.__bases__
        )
        self.__dict__.update(instance_data)



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest
import skhep_testdata

import uproot


def test_version():
    assert uproot.classname_decode(
        uproot.classname_encode("xAOD::MissingETAuxAssociationMap_v2")
    ) == ("xAOD::MissingETAuxAssociationMap_v2", None)
    assert uproot.classname_decode(
        uproot.classname_encode("xAOD::MissingETAuxAssociationMap_v2", 9)
    ) == ("xAOD::MissingETAuxAssociationMap_v2", 9)



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys

import numpy
import pytest
import skhep_testdata

import uproot
from uproot.containers import AsMap, AsSet, AsString, AsVector
from uproot.interpretation.jagged import AsJagged
from uproot.interpretation.numerical import AsDtype
from uproot.interpretation.objects import AsObjects


@pytest.mark.skip(
    reason="Implement non-memberwise std::map; we have a sample (map<string,double>)"
)
def test_nonmemberwise_asmap():
    with uproot.open(skhep_testdata.data_path("uproot-issue243.root")) as file:
        branch = file["triggerList/triggerMap"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(library="np", entry_stop=2)

    with uproot.open(skhep_testdata.data_path("uproot-issue-268.root")) as file:
        branch = file["aTree/VtxTracks/channels_"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(library="np", entry_stop=2)

    with uproot.open(skhep_testdata.data_path("uproot-issue-268.root")) as file:
        branch = file["aTree/VtxTracks2TofHits/match_"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(library="np", entry_stop=2)

    with uproot.open(skhep_testdata.data_path("uproot-issue-268.root")) as file:
        branch = file["aTree/VtxTracks2TofHits/match_inverted_"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(library="np", entry_stop=2)


def test_typename():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert tree["vector_int32"].interpretation == AsJagged(AsDtype(">i4"), 10)
        assert tree["vector_string"].interpretation == AsObjects(
            AsVector(True, AsString(False))
        )
        assert tree["vector_vector_int32"].interpretation == AsObjects(
            AsVector(True, AsVector(False, numpy.dtype(">i4")))
        )
        assert tree["vector_vector_string"].interpretation == AsObjects(
            AsVector(True, AsVector(False, AsString(False)))
        )
        assert tree["vector_set_int32"].interpretation == AsObjects(
            AsVector(True, AsSet(False, numpy.dtype(">i4")))
        )
        assert tree["vector_set_string"].interpretation == AsObjects(
            AsVector(True, AsSet(False, AsString(False)))
        )
        assert tree["set_int32"].interpretation == AsObjects(
            AsSet(True, numpy.dtype(">i4"))
        )
        assert tree["set_string"].interpretation == AsObjects(
            AsSet(True, AsString(False))
        )
        assert tree["map_int32_int16"].interpretation == AsObjects(
            AsMap(True, numpy.dtype(">i4"), numpy.dtype(">i2"))
        )
        assert tree["map_int32_vector_int16"].interpretation == AsObjects(
            AsMap(True, numpy.dtype(">i4"), AsVector(True, numpy.dtype(">i2")))
        )
        assert tree["map_int32_vector_string"].interpretation == AsObjects(
            AsMap(True, numpy.dtype(">i4"), AsVector(True, AsString(False)))
        )
        assert tree["map_int32_set_int16"].interpretation == AsObjects(
            AsMap(True, numpy.dtype(">i4"), AsSet(True, numpy.dtype(">i2")))
        )
        assert tree["map_int32_set_string"].interpretation == AsObjects(
            AsMap(True, numpy.dtype(">i4"), AsSet(True, AsString(False)))
        )
        assert tree["map_string_int16"].interpretation == AsObjects(
            AsMap(True, AsString(True), numpy.dtype(">i2"))
        )
        assert tree["map_string_vector_int16"].interpretation == AsObjects(
            AsMap(True, AsString(True), AsVector(True, numpy.dtype(">i2")))
        )
        assert tree["map_string_vector_string"].interpretation == AsObjects(
            AsMap(True, AsString(True), AsVector(True, AsString(False)))
        )
        assert tree["map_string_set_int16"].interpretation == AsObjects(
            AsMap(True, AsString(True), AsSet(True, numpy.dtype(">i2")))
        )
        assert tree["map_string_set_string"].interpretation == AsObjects(
            AsMap(True, AsString(True), AsSet(True, AsString(False)))
        )
        assert tree["map_int32_vector_vector_int16"].interpretation == AsObjects(
            AsMap(
                True,
                numpy.dtype(">i4"),
                AsVector(True, AsVector(False, numpy.dtype(">i2"))),
            )
        )
        assert tree["map_int32_vector_set_int16"].interpretation == AsObjects(
            AsMap(
                True,
                numpy.dtype(">i4"),
                AsVector(True, AsSet(False, numpy.dtype(">i2"))),
            )
        )
        assert tree["map_string_string"].interpretation == AsObjects(
            AsMap(True, AsString(True), AsString(True))
        )


def test_string():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert tree["string"].array(library="np").tolist() == [
            "one",
            "two",
            "three",
            "four",
            "five",
        ]


def test_tstring():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert tree["tstring"].array(library="np").tolist() == [
            "one",
            "two",
            "three",
            "four",
            "five",
        ]


def test_vector_int32():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["vector_int32"].array(library="np")] == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]


def test_vector_string():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["vector_string"].array(library="np")] == [
            ["one"],
            ["one", "two"],
            ["one", "two", "three"],
            ["one", "two", "three", "four"],
            ["one", "two", "three", "four", "five"],
        ]


def test_vector_tstring():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["vector_tstring"].array(library="np")] == [
            ["one"],
            ["one", "two"],
            ["one", "two", "three"],
            ["one", "two", "three", "four"],
            ["one", "two", "three", "four", "five"],
        ]


def test_vector_vector_int32():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["vector_vector_int32"].array(library="np")
        ] == [
            [[1]],
            [[1], [1, 2]],
            [[1], [1, 2], [1, 2, 3]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]],
        ]


def test_vector_vector_string():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["vector_vector_string"].array(library="np")
        ] == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "two"], ["one", "two", "three"]],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
            ],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
                ["one", "two", "three", "four", "five"],
            ],
        ]


def test_vector_set_int32():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["vector_set_int32"].array(library="np")] == [
            [{1}],
            [{1}, {1, 2}],
            [{1}, {1, 2}, {1, 2, 3}],
            [{1}, {1, 2}, {1, 2, 3}, {1, 2, 3, 4}],
            [
                {1},
                {1, 2},
                {1, 2, 3},
                {1, 2, 3, 4},
                {1, 2, 3, 4, 5},
            ],
        ]


def test_vector_set_string():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["vector_set_string"].array(library="np")] == [
            [{"one"}],
            [{"one"}, {"one", "two"}],
            [{"one"}, {"one", "two"}, {"one", "two", "three"}],
            [
                {"one"},
                {"one", "two"},
                {"one", "two", "three"},
                {"one", "two", "three", "four"},
            ],
            [
                {"one"},
                {"one", "two"},
                {"one", "two", "three"},
                {"one", "two", "three", "four"},
                {"one", "two", "three", "four", "five"},
            ],
        ]


def test_set_int32():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["set_int32"].array(library="np")] == [
            {1},
            {1, 2},
            {1, 2, 3},
            {1, 2, 3, 4},
            {1, 2, 3, 4, 5},
        ]


def test_set_string():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["set_string"].array(library="np")] == [
            {"one"},
            {"one", "two"},
            {"one", "two", "three"},
            {"one", "two", "three", "four"},
            {"one", "two", "three", "four", "five"},
        ]


def test_map_int32_int16():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["map_int32_int16"].array(library="np")] == [
            {1: 1},
            {1: 1, 2: 2},
            {1: 1, 2: 2, 3: 3},
            {1: 1, 2: 2, 3: 3, 4: 4},
            {1: 1, 2: 2, 3: 3, 4: 4, 5: 5},
        ]


def test_map_int32_vector_int16():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["map_int32_vector_int16"].array(library="np")
        ] == [
            {1: [1]},
            {1: [1], 2: [1, 2]},
            {1: [1], 2: [1, 2], 3: [1, 2, 3]},
            {1: [1], 2: [1, 2], 3: [1, 2, 3], 4: [1, 2, 3, 4]},
            {1: [1], 2: [1, 2], 3: [1, 2, 3], 4: [1, 2, 3, 4], 5: [1, 2, 3, 4, 5]},
        ]


def test_map_int32_vector_string():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["map_int32_vector_string"].array(library="np")
        ] == [
            {1: ["one"]},
            {1: ["one"], 2: ["one", "two"]},
            {1: ["one"], 2: ["one", "two"], 3: ["one", "two", "three"]},
            {
                1: ["one"],
                2: ["one", "two"],
                3: ["one", "two", "three"],
                4: ["one", "two", "three", "four"],
            },
            {
                1: ["one"],
                2: ["one", "two"],
                3: ["one", "two", "three"],
                4: ["one", "two", "three", "four"],
                5: ["one", "two", "three", "four", "five"],
            },
        ]


def test_map_int32_set_int16():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["map_int32_set_int16"].array(library="np")
        ] == [
            {1: {1}},
            {1: {1}, 2: {1, 2}},
            {1: {1}, 2: {1, 2}, 3: {1, 2, 3}},
            {1: {1}, 2: {1, 2}, 3: {1, 2, 3}, 4: {1, 2, 3, 4}},
            {
                1: {1},
                2: {1, 2},
                3: {1, 2, 3},
                4: {1, 2, 3, 4},
                5: {1, 2, 3, 4, 5},
            },
        ]


def test_map_int32_set_string():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["map_int32_set_string"].array(library="np")
        ] == [
            {1: {"one"}},
            {1: {"one"}, 2: {"one", "two"}},
            {1: {"one"}, 2: {"one", "two"}, 3: {"one", "two", "three"}},
            {
                1: {"one"},
                2: {"one", "two"},
                3: {"one", "two", "three"},
                4: {"one", "two", "three", "four"},
            },
            {
                1: {"one"},
                2: {"one", "two"},
                3: {"one", "two", "three"},
                4: {"one", "two", "three", "four"},
                5: {"one", "two", "three", "four", "five"},
            },
        ]


def test_map_string_int16():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["map_string_int16"].array(library="np")] == [
            {"one": 1},
            {"one": 1, "two": 2},
            {"one": 1, "two": 2, "three": 3},
            {"one": 1, "two": 2, "three": 3, "four": 4},
            {"one": 1, "two": 2, "three": 3, "four": 4, "five": 5},
        ]


def test_map_string_vector_int16():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["map_string_vector_int16"].array(library="np")
        ] == [
            {"one": [1]},
            {"one": [1], "two": [1, 2]},
            {"one": [1], "two": [1, 2], "three": [1, 2, 3]},
            {"one": [1], "two": [1, 2], "three": [1, 2, 3], "four": [1, 2, 3, 4]},
            {
                "one": [1],
                "two": [1, 2],
                "three": [1, 2, 3],
                "four": [1, 2, 3, 4],
                "five": [1, 2, 3, 4, 5],
            },
        ]


def test_map_string_vector_string():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["map_string_vector_string"].array(library="np")
        ] == [
            {"one": ["one"]},
            {"one": ["one"], "two": ["one", "two"]},
            {"one": ["one"], "two": ["one", "two"], "three": ["one", "two", "three"]},
            {
                "one": ["one"],
                "two": ["one", "two"],
                "three": ["one", "two", "three"],
                "four": ["one", "two", "three", "four"],
            },
            {
                "one": ["one"],
                "two": ["one", "two"],
                "three": ["one", "two", "three"],
                "four": ["one", "two", "three", "four"],
                "five": ["one", "two", "three", "four", "five"],
            },
        ]


def test_map_string_set_int16():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["map_string_set_int16"].array(library="np")
        ] == [
            {"one": {1}},
            {"one": {1}, "two": {1, 2}},
            {"one": {1}, "two": {1, 2}, "three": {1, 2, 3}},
            {
                "one": {1},
                "two": {1, 2},
                "three": {1, 2, 3},
                "four": {1, 2, 3, 4},
            },
            {
                "one": {1},
                "two": {1, 2},
                "three": {1, 2, 3},
                "four": {1, 2, 3, 4},
                "five": {1, 2, 3, 4, 5},
            },
        ]


def test_map_string_set_string():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["map_string_set_string"].array(library="np")
        ] == [
            {"one": {"one"}},
            {"one": {"one"}, "two": {"one", "two"}},
            {
                "one": {"one"},
                "two": {"one", "two"},
                "three": {"one", "two", "three"},
            },
            {
                "one": {"one"},
                "two": {"one", "two"},
                "three": {"one", "two", "three"},
                "four": {"one", "two", "three", "four"},
            },
            {
                "one": {"one"},
                "two": {"one", "two"},
                "three": {"one", "two", "three"},
                "four": {"one", "two", "three", "four"},
                "five": {"one", "two", "three", "four", "five"},
            },
        ]


def test_map_int32_vector_vector_int16():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist()
            for x in tree["map_int32_vector_vector_int16"].array(library="np")
        ] == [
            {1: [[1]]},
            {1: [[1]], 2: [[1], [1, 2]]},
            {1: [[1]], 2: [[1], [1, 2]], 3: [[1], [1, 2], [1, 2, 3]]},
            {
                1: [[1]],
                2: [[1], [1, 2]],
                3: [[1], [1, 2], [1, 2, 3]],
                4: [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
            },
            {
                1: [[1]],
                2: [[1], [1, 2]],
                3: [[1], [1, 2], [1, 2, 3]],
                4: [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
                5: [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]],
            },
        ]


def test_map_int32_vector_set_int16():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [
            x.tolist() for x in tree["map_int32_vector_set_int16"].array(library="np")
        ] == [
            {1: [{1}]},
            {1: [{1}], 2: [{1}, {1, 2}]},
            {
                1: [{1}],
                2: [{1}, {1, 2}],
                3: [{1}, {1, 2}, {1, 2, 3}],
            },
            {
                1: [{1}],
                2: [{1}, {1, 2}],
                3: [{1}, {1, 2}, {1, 2, 3}],
                4: [{1}, {1, 2}, {1, 2, 3}, {1, 2, 3, 4}],
            },
            {
                1: [{1}],
                2: [{1}, {1, 2}],
                3: [{1}, {1, 2}, {1, 2, 3}],
                4: [{1}, {1, 2}, {1, 2, 3}, {1, 2, 3, 4}],
                5: [
                    {1},
                    {1, 2},
                    {1, 2, 3},
                    {1, 2, 3, 4},
                    {1, 2, 3, 4, 5},
                ],
            },
        ]


def test_map_string_string():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["map_string_string"].array(library="np")] == [
            {"one": "ONE"},
            {"one": "ONE", "two": "TWO"},
            {"one": "ONE", "two": "TWO", "three": "THREE"},
            {"one": "ONE", "two": "TWO", "three": "THREE", "four": "FOUR"},
            {
                "one": "ONE",
                "two": "TWO",
                "three": "THREE",
                "four": "FOUR",
                "five": "FIVE",
            },
        ]


def test_map_string_tstring():
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert [x.tolist() for x in tree["map_string_tstring"].array(library="np")] == [
            {"one": "ONE"},
            {"one": "ONE", "two": "TWO"},
            {"one": "ONE", "two": "TWO", "three": "THREE"},
            {"one": "ONE", "two": "TWO", "three": "THREE", "four": "FOUR"},
            {
                "one": "ONE",
                "two": "TWO",
                "three": "THREE",
                "four": "FOUR",
                "five": "FIVE",
            },
        ]


def test_map_int_struct():
    with uproot.open(skhep_testdata.data_path("uproot-issue468.root"))[
        "Geant4Data/Geant4Data./Geant4Data.particles"
    ] as branch:
        assert (
            repr(branch.interpretation) == "AsObjects(AsMap(True, dtype('>i4'), "
            "Model_BDSOutputROOTGeant4Data_3a3a_ParticleInfo))"
        )
        result = branch.array(library="np")[0]
        assert result.keys().tolist() == [
            -1000020040,
            -1000020030,
            -1000010030,
            -1000010020,
            -2212,
            -2112,
            -321,
            -211,
            -16,
            -15,
            -14,
            -13,
            -12,
            -11,
            0,
            11,
            12,
            13,
            14,
            15,
            16,
            22,
            211,
            321,
            2112,
            2212,
            1000010020,
            1000010030,
            1000020030,
            1000020040,
        ]
        assert [x.member("name") for x in result.values()] == [
            "anti_alpha",
            "anti_He3",
            "anti_triton",
            "anti_deuteron",
            "anti_proton",
            "anti_neutron",
            "kaon-",
            "pi-",
            "anti_nu_tau",
            "tau+",
            "anti_nu_mu",
            "mu+",
            "anti_nu_e",
            "e+",
            "geantino",
            "e-",
            "nu_e",
            "mu-",
            "nu_mu",
            "tau-",
            "nu_tau",
            "gamma",
            "pi+",
            "kaon+",
            "neutron",
            "proton",
            "deuteron",
            "triton",
            "He3",
            "alpha",
        ]
        assert [x.member("charge") for x in result.values()] == [
            -2,
            -2,
            -1,
            -1,
            -1,
            0,
            -1,
            -1,
            0,
            1,
            0,
            1,
            0,
            1,
            0,
            -1,
            0,
            -1,
            0,
            -1,
            0,
            0,
            1,
            1,
            0,
            1,
            1,
            1,
            2,
            2,
        ]
        assert [x.member("mass") for x in result.values()] == [
            3.727379,
            2.808391,
            2.808921,
            1.875613,
            0.938272013,
            0.93956536,
            0.493677,
            0.1395701,
            0.0,
            1.77686,
            0.0,
            0.1056583715,
            0.0,
            0.00051099891,
            0.0,
            0.00051099891,
            0.0,
            0.1056583715,
            0.0,
            1.77686,
            0.0,
            0.0,
            0.1395701,
            0.493677,
            0.93956536,
            0.938272013,
            1.875613,
            2.808921,
            2.808391,
            3.727379,
        ]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import numpy as np
import pytest
import skhep_testdata

import uproot

hist = pytest.importorskip("hist")


def test_no_flow(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    h = hist.Hist(hist.axis.Regular(20, 0, 20, flow=False)).fill(
        np.random.normal(10, 6, 1000)
    )

    with uproot.recreate(filename) as fout:
        fout["test"] = h

    with uproot.open(filename) as fin:
        h_read = fin["test"]

    assert np.array_equal(h.axes[0].edges, h_read.axes[0].edges())
    assert np.array_equal(h.values(flow=False), h_read.values(flow=False))


def test_yes_flow(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    h = hist.Hist(hist.axis.Regular(20, 0, 20, flow=True)).fill(
        np.random.normal(10, 6, 1000)
    )

    with uproot.recreate(filename) as fout:
        fout["test"] = h

    with uproot.open(filename) as fin:
        h_read = fin["test"]

    assert np.array_equal(h.axes[0].edges, h_read.axes[0].edges())
    assert np.array_equal(h.values(flow=False), h_read.values(flow=False))
    assert np.array_equal(h.values(flow=True), h_read.values(flow=True))



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import pytest

import uproot
import uproot.writing


def test_write_empty(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    tlist = uproot.writing.identify.to_TList([])

    with uproot.recreate(filename) as f:
        f["test"] = tlist

    with uproot.open(filename) as f2:
        assert len(f2["test"]) == 0


def test_write_single_key(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    tlist = uproot.writing.identify.to_TList(
        [uproot.writing.identify.to_TObjString("test string")]
    )

    with uproot.recreate(filename) as f:
        f["test"] = tlist

    with uproot.open(filename) as f2:
        assert len(f2["test"]) == 1


def test_write_nested(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    tlist_child = uproot.writing.identify.to_TList(
        [uproot.writing.identify.to_TObjString(s) for s in "this is a test".split()]
    )

    entries = [
        uproot.writing.identify.to_TObjString("this string goes in the front"),
        tlist_child,
        uproot.writing.identify.to_TObjString("test string"),
    ]

    tlist = uproot.writing.identify.to_TList(entries)

    with uproot.recreate(filename) as f:
        f["test"] = tlist

    with uproot.open(filename) as f2:
        parent_list = f2["test"]
        assert len(parent_list) == 3
        assert isinstance(parent_list[0], uproot.models.TObjString.Model_TObjString)
        assert str(parent_list[0]) == "this string goes in the front"
        assert str(parent_list[2]) == "test string"
        child_list = parent_list[1]
        assert isinstance(child_list, uproot.models.TList.Model_TList)
        assert len(child_list) == 4
        assert " ".join([str(s) for s in child_list]) == "this is a test"



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines behaviors, which are mix-in classes that provide a high-level interface
to objects read from ROOT files.

Behaviors do not have a required base class, and they may be used with
Awkward Array's ``ak.behavior``.

To add a behavior for a ROOT class:

1. Translate the ROOT class name from C++ to Python with
   :doc:`uproot.model.classname_encode`. For example,
   ``"ROOT::RThing"`` becomes ``"Model_ROOT_3a3a_RThing"``.
2. Create a submodule of ``uproot.behaviors`` without
   the ``"Model_"`` prefix. For example, ``"ROOT_3a3a_RThing"``.
3. Include a class in that submodule with the fully encoded
   name. For example, ``"Model_ROOT_3a3a_RThing"``.

When Uproot reads an instance of the class, it would always create a
deserialization model if one is not predefined. But if a behavior with the
appropriate name exist, the new class will inherit from the behavior, giving
the newly created object specialized methods and properties.

See also :doc:`uproot.models`.
"""



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import array
import os

import numpy as np
import pytest
import skhep_testdata

import uproot
import uproot.writing

ROOT = pytest.importorskip("ROOT")
awkward = pytest.importorskip("awkward")


def test_awkward_as_numpy(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout.mktree(
            "tree",
            {
                "b1": awkward.types.from_datashape("int32", highlevel=False),
                "b2": awkward.types.from_datashape("2 * float64", highlevel=False),
                "b3": awkward.types.from_datashape("2 * 3 * float64", highlevel=False),
                "b4": awkward.Array([1.1, 2.2, 3.3]).type,
            },
        )

    with uproot.open(newfile) as fin:
        assert fin["tree/b1"].typename == "int32_t"
        assert fin["tree/b2"].typename == "double[2]"
        assert fin["tree/b3"].typename == "double[2][3]"
        assert fin["tree/b4"].typename == "double"


def test_awkward_record(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout.mktree(
            "tree",
            {
                "b1": "int32",
                "b2": awkward.types.from_datashape(
                    '{"x": float64, "y": 3 * float64}', highlevel=False
                ),
            },
        )

    with uproot.open(newfile) as fin:
        assert fin["tree/b1"].typename == "int32_t"
        assert fin["tree/b2_x"].typename == "double"
        assert fin["tree/b2_y"].typename == "double[3]"


def test_awkward_record_data(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        b1 = np.array([1, 2, 3], np.int32)
        b2 = awkward.Array([{"x": 1.1, "y": 4}, {"x": 2.2, "y": 5}, {"x": 3.3, "y": 6}])
        fout.mktree("tree", {"b1": b1.dtype, "b2": b2.type})
        fout["tree"].extend({"b1": b1, "b2": b2})

    with uproot.open(newfile) as fin:
        assert fin["tree/b1"].typename == "int32_t"
        assert fin["tree/b2_x"].typename == "double"
        assert fin["tree/b2_y"].typename == "int64_t"
        assert fin["tree/b1"].array().tolist() == [1, 2, 3]
        assert fin["tree/b2_x"].array().tolist() == [1.1, 2.2, 3.3]
        assert fin["tree/b2_y"].array().tolist() == [4, 5, 6]


def test_awkward_record_dict_1(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        b1 = np.array([1, 2, 3], np.int32)
        b2 = {"x": np.array([1.1, 2.2, 3.3]), "y": np.array([4, 5, 6], np.int64)}
        fout.mktree(
            "tree", {"b1": b1.dtype, "b2": {"x": b2["x"].dtype, "y": b2["y"].dtype}}
        )
        fout["tree"].extend({"b1": b1, "b2": b2})

    with uproot.open(newfile) as fin:
        assert fin["tree/b1"].typename == "int32_t"
        assert fin["tree/b2_x"].typename == "double"
        assert fin["tree/b2_y"].typename == "int64_t"
        assert fin["tree/b1"].array().tolist() == [1, 2, 3]
        assert fin["tree/b2_x"].array().tolist() == [1.1, 2.2, 3.3]
        assert fin["tree/b2_y"].array().tolist() == [4, 5, 6]


def test_awkward_record_dict_2(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        b1 = np.array([1, 2, 3], np.int32)
        b2 = {"x": np.array([1.1, 2.2, 3.3]), "y": np.array([4, 5, 6], np.int64)}
        fout["tree"] = {"b1": b1, "b2": b2}
        fout["tree"].extend({"b1": b1, "b2": b2})

    with uproot.open(newfile) as fin:
        assert fin["tree/b1"].typename == "int32_t"
        assert fin["tree/b2_x"].typename == "double"
        assert fin["tree/b2_y"].typename == "int64_t"
        assert fin["tree/b1"].array().tolist() == [1, 2, 3, 1, 2, 3]
        assert fin["tree/b2_x"].array().tolist() == [1.1, 2.2, 3.3, 1.1, 2.2, 3.3]
        assert fin["tree/b2_y"].array().tolist() == [4, 5, 6, 4, 5, 6]


def test_awkward_record_recarray(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        b1 = np.array([1, 2, 3], np.int32)
        b2 = np.array(
            [(1.1, 4), (2.2, 5), (3.3, 6)], [("x", np.float64), ("y", np.int64)]
        )
        fout["tree"] = {"b1": b1, "b2": b2}
        fout["tree"].extend({"b1": b1, "b2": b2})

    with uproot.open(newfile) as fin:
        assert fin["tree/b1"].typename == "int32_t"
        assert fin["tree/b2_x"].typename == "double"
        assert fin["tree/b2_y"].typename == "int64_t"
        assert fin["tree/b1"].array().tolist() == [1, 2, 3, 1, 2, 3]
        assert fin["tree/b2_x"].array().tolist() == [1.1, 2.2, 3.3, 1.1, 2.2, 3.3]
        assert fin["tree/b2_y"].array().tolist() == [4, 5, 6, 4, 5, 6]


def test_awkward_record_pandas(tmp_path):
    pandas = pytest.importorskip("pandas")

    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        b1 = np.array([1, 2, 3], np.int32)
        b2 = pandas.DataFrame({"x": [1.1, 2.2, 3.3], "y": [4, 5, 6]})
        fout["tree"] = {"b1": b1, "b2": b2}
        fout["tree"].extend({"b1": b1, "b2": b2})

    with uproot.open(newfile) as fin:
        assert fin["tree/b1"].typename == "int32_t"
        assert fin["tree/b2_x"].typename == "double"
        assert fin["tree/b2_y"].typename == "int64_t"
        assert fin["tree/b1"].array().tolist() == [1, 2, 3, 1, 2, 3]
        assert fin["tree/b2_x"].array().tolist() == [1.1, 2.2, 3.3, 1.1, 2.2, 3.3]
        assert fin["tree/b2_y"].array().tolist() == [4, 5, 6, 4, 5, 6]


def test_top_level(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    df1 = awkward.Array({"x": [1, 2, 3], "y": [1.1, 2.2, 3.3]})
    df2 = awkward.Array({"x": [4, 5, 6], "y": [4.4, 5.5, 6.6]})

    with uproot.recreate(newfile, compression=None) as fout:
        fout["tree"] = df1
        fout["tree"].extend(df2)

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")

    assert t1.GetBranch("x").GetName() == "x"
    assert t1.GetBranch("y").GetName() == "y"
    assert [np.asarray(x.x).tolist() for x in t1] == [1, 2, 3, 4, 5, 6]
    assert [np.asarray(x.y).tolist() for x in t1] == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]

    with uproot.open(newfile) as fin:
        assert fin["tree/x"].name == "x"
        assert fin["tree/y"].name == "y"
        assert fin["tree/x"].typename.startswith("int")
        assert fin["tree/y"].typename == "double"
        assert fin["tree/x"].array().tolist() == [1, 2, 3, 4, 5, 6]
        assert fin["tree/y"].array().tolist() == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]

    f1.Close()


def test_awkward_jagged_metadata(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout.mktree(
            "tree",
            {
                "b1": "int64",
                "b2": awkward.types.from_datashape("var * float64", highlevel=False),
            },
        )

    with uproot.open(newfile) as fin:
        assert fin["tree/b1"].typename == "int64_t"
        assert fin["tree/nb2"].typename == "int32_t"
        assert fin["tree/b2"].typename == "double[]"

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")

    b1 = t1.GetBranch("b1")
    assert b1.GetLeaf("b1").GetName() == "b1"
    assert not b1.GetLeaf("b1").GetLeafCount()

    nb2 = t1.GetBranch("nb2")
    assert nb2.GetLeaf("nb2").GetName() == "nb2"
    assert not nb2.GetLeaf("nb2").GetLeafCount()

    b2 = t1.GetBranch("b2")
    assert b2.GetLeaf("b2").GetName() == "b2"
    assert b2.GetLeaf("b2").GetLeafCount()
    assert b2.GetLeaf("b2").GetLeafCount().GetName() == "nb2"

    f1.Close()


def test_awkward_jagged_record_metadata(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout.mktree(
            "tree",
            {
                "b1": "int64",
                "b2": awkward.types.from_datashape(
                    'var * {"x": float64, "y": int8}', highlevel=False
                ),
            },
        )

    with uproot.open(newfile) as fin:
        assert fin["tree/b1"].typename == "int64_t"
        assert fin["tree/nb2"].typename == "int32_t"
        assert fin["tree/b2_x"].typename == "double[]"
        assert fin["tree/b2_y"].typename == "int8_t[]"

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")

    b1 = t1.GetBranch("b1")
    assert b1.GetLeaf("b1").GetName() == "b1"
    assert not b1.GetLeaf("b1").GetLeafCount()

    nb2 = t1.GetBranch("nb2")
    assert nb2.GetLeaf("nb2").GetName() == "nb2"
    assert not nb2.GetLeaf("nb2").GetLeafCount()

    b2_x = t1.GetBranch("b2_x")
    assert b2_x.GetLeaf("b2_x").GetName() == "b2_x"
    assert b2_x.GetLeaf("b2_x").GetLeafCount()
    assert b2_x.GetLeaf("b2_x").GetLeafCount().GetName() == "nb2"

    b2_y = t1.GetBranch("b2_y")
    assert b2_y.GetLeaf("b2_y").GetName() == "b2_y"
    assert b2_y.GetLeaf("b2_y").GetLeafCount()
    assert b2_y.GetLeaf("b2_y").GetLeafCount().GetName() == "nb2"

    f1.Close()


def test_awkward_jagged_data_1(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        b1 = np.array([1, 2, 3, 4, 5], np.int64)
        b2 = awkward.Array(
            [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
        )
        fout.mktree("tree", {"b1": b1.dtype, "b2": b2.type})
        fout["tree"].extend({"b1": b1, "b2": b2})

    with uproot.open(newfile) as fin:
        assert fin["tree/nb2"].member("fLeaves")[0].member("fMaximum") == 4
        assert fin["tree/b2"].member("fEntryOffsetLen") == 4 * 5
        assert fin["tree/b1"].array().tolist() == [1, 2, 3, 4, 5]
        assert fin["tree/nb2"].array().tolist() == [3, 0, 2, 1, 4]
        assert fin["tree/b2"].array().tolist() == [
            [0.0, 1.1, 2.2],
            [],
            [3.3, 4.4],
            [5.5],
            [6.6, 7.7, 8.8, 9.9],
        ]

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")
    assert [x.b1 for x in t1] == [1, 2, 3, 4, 5]
    assert [x.nb2 for x in t1] == [3, 0, 2, 1, 4]
    assert [list(x.b2) for x in t1] == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    f1.Close()


def test_awkward_jagged_data_2(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        b1 = np.array([1, 2, 3, 4, 5], np.int64)
        b2 = awkward.Array(
            [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
        )
        fout["tree"] = {"b1": b1, "b2": b2}
        fout["tree"].extend({"b1": b1[:3], "b2": b2[:3]})

    with uproot.open(newfile) as fin:
        assert fin["tree/nb2"].member("fLeaves")[0].member("fMaximum") == 4
        assert fin["tree/b2"].member("fEntryOffsetLen") == 4 * 3
        assert fin["tree/b1"].array().tolist() == [1, 2, 3, 4, 5, 1, 2, 3]
        assert fin["tree/nb2"].array().tolist() == [3, 0, 2, 1, 4, 3, 0, 2]
        assert fin["tree/b2"].array().tolist() == [
            [0.0, 1.1, 2.2],
            [],
            [3.3, 4.4],
            [5.5],
            [6.6, 7.7, 8.8, 9.9],
            [0.0, 1.1, 2.2],
            [],
            [3.3, 4.4],
        ]

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")
    assert [x.b1 for x in t1] == [1, 2, 3, 4, 5, 1, 2, 3]
    assert [x.nb2 for x in t1] == [3, 0, 2, 1, 4, 3, 0, 2]
    assert [list(x.b2) for x in t1] == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
    ]
    f1.Close()


def test_awkward_jagged_data_3(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        big = awkward.Array(
            [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]] * 300
        )
        fout["tree"] = {"big": big}
        # more than 1000 entries, a special number for fNevBufSize and fEntryOffsetLen

    with uproot.open(newfile) as fin:
        assert fin["tree/nbig"].member("fLeaves")[0].member("fMaximum") == 4
        assert fin["tree/big"].member("fEntryOffsetLen") == 4 * 1500
        assert fin["tree/nbig"].array().tolist() == [3, 0, 2, 1, 4] * 300
        assert (
            fin["tree/big"].array().tolist()
            == [
                [0.0, 1.1, 2.2],
                [],
                [3.3, 4.4],
                [5.5],
                [6.6, 7.7, 8.8, 9.9],
            ]
            * 300
        )

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")
    assert [x.nbig for x in t1] == [3, 0, 2, 1, 4] * 300
    assert [list(x.big) for x in t1] == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ] * 300
    f1.Close()


def test_awkward_jagged_record_1(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        array = awkward.Array(
            [
                [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
                [],
                [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
            ]
        )
        fout["tree"] = {"array": array}
        fout["tree"].extend({"array": array})

    with uproot.open(newfile) as fin:
        assert fin["tree/narray"].array().tolist() == [3, 0, 2] * 2
        assert fin["tree/array_x"].array().tolist() == [[1, 2, 3], [], [4, 5]] * 2
        assert (
            fin["tree/array_y"].array().tolist()
            == [[1.1, 2.2, 3.3], [], [4.4, 5.5]] * 2
        )

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")
    assert [x.narray for x in t1] == [3, 0, 2] * 2
    assert [list(x.array_x) for x in t1] == [[1, 2, 3], [], [4, 5]] * 2
    assert [list(x.array_y) for x in t1] == [[1.1, 2.2, 3.3], [], [4.4, 5.5]] * 2
    f1.Close()


def test_awkward_jagged_record_2(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout["tree"] = awkward.Array(
            [
                [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
                [],
                [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
            ]
        )
        fout["tree"].extend(
            awkward.Array(
                [
                    [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
                    [],
                    [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
                ]
            )
        )

    with uproot.open(newfile) as fin:
        assert fin["tree/n"].array().tolist() == [3, 0, 2] * 2
        assert fin["tree/x"].array().tolist() == [[1, 2, 3], [], [4, 5]] * 2
        assert fin["tree/y"].array().tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]] * 2

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")
    assert [x.n for x in t1] == [3, 0, 2] * 2
    assert [list(x.x) for x in t1] == [[1, 2, 3], [], [4, 5]] * 2
    assert [list(x.y) for x in t1] == [[1.1, 2.2, 3.3], [], [4.4, 5.5]] * 2
    f1.Close()



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import skhep_testdata

import uproot


def test_num_entries_single():
    file1 = skhep_testdata.data_path("uproot-Zmumu.root")
    ttree1 = file1 + ":events"
    assert list(uproot.num_entries(ttree1)) == [(file1, "events", 2304)]


def test_num_entries_multiple():
    file1 = skhep_testdata.data_path("uproot-Zmumu.root")
    ttree1 = file1 + ":events"
    file2 = skhep_testdata.data_path("uproot-HZZ.root")
    ttree2 = file2 + ":events"
    file3 = skhep_testdata.data_path("uproot-sample-6.08.04-uncompressed.root")
    ttree3 = file3 + ":sample"

    assert list(uproot.num_entries([ttree1, ttree2, ttree3])) == [
        (file1, "events", 2304),
        (file2, "events", 2421),
        (file3, "sample", 30),
    ]


def test_num_entries_as_iterator():
    file1 = skhep_testdata.data_path("uproot-Zmumu.root")
    ttree1 = file1 + ":events"
    file2 = skhep_testdata.data_path("uproot-HZZ.root")
    ttree2 = file2 + ":events"
    file3 = skhep_testdata.data_path("uproot-sample-6.08.04-uncompressed.root")
    ttree3 = file3 + ":sample"

    vals = [
        (file1, "events", 2304),
        (file2, "events", 2421),
        (file3, "sample", 30),
    ]
    for i, num in enumerate(uproot.num_entries([ttree1, ttree2, ttree3])):
        assert num == vals[i]


def test_dict_input():
    file1 = skhep_testdata.data_path("uproot-Zmumu.root")
    file2 = skhep_testdata.data_path("uproot-HZZ.root")
    file3 = skhep_testdata.data_path("uproot-sample-6.08.04-uncompressed.root")

    vals = [
        (file1, "events", 2304),
        (file2, "events", 2421),
        (file3, "sample", 30),
    ]
    for i, num in enumerate(
        uproot.num_entries({file1: "events", file2: "events", file3: "sample"})
    ):
        assert num == vals[i]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os
import shutil
import struct

import numpy as np
import pytest
import skhep_testdata

import uproot

ROOT = pytest.importorskip("ROOT")


_file_header_fields_small = struct.Struct(">4siiiiiiiBiiiH16s")
_file_header_fields_big = struct.Struct(">4siiqqiiiBiqiH16s")

_directory_format_small = struct.Struct(">hIIiiiii")
_directory_format_big = struct.Struct(">hIIiiqqq")
_directory_format_num_keys = struct.Struct(">i")

_key_format_small = struct.Struct(">ihiIhhii")
_key_format_big = struct.Struct(">ihiIhhqq")

_string_size_format_1 = struct.Struct(">B")
_string_size_format_4 = struct.Struct(">I")

_free_format_small = struct.Struct(">HII")
_free_format_big = struct.Struct(">HQQ")

# /home/jpivarski/storage/data/uproot4-big/issue-288.root
# /home/jpivarski/storage/data/uproot4-big/issue-90.root


@pytest.mark.parametrize(
    "source", ["uproot-Zmumu.root", "uproot-issue243-new.root", "uproot-issue64.root"]
)
def test(tmp_path, source):
    # print(source)

    filename = os.path.join(tmp_path, "update-me.root")
    shutil.copyfile(skhep_testdata.data_path(source), filename)

    # print("BEFORE")
    # for version, fFirst, fLast in get_frees(filename):
    #     print(version, fFirst, fLast)

    with open(filename, "r+b") as rawfile:
        rawfile.seek(0)
        (
            magic,
            fVersion,
            fBEGIN,
            fEND,
            fSeekFree,
            fNbytesFree,
            nfree,
            fNbytesName,
            fUnits,
            fCompress,
            fSeekInfo,
            fNbytesInfo,
            fUUID_version,
            fUUID,
        ) = _file_header_fields_small.unpack(
            rawfile.read(_file_header_fields_small.size)
        )

        if fVersion > 1000000:
            rawfile.seek(0)
            (
                magic,
                fVersion,
                fBEGIN,
                fEND,
                fSeekFree,
                fNbytesFree,
                nfree,
                fNbytesName,
                fUnits,
                fCompress,
                fSeekInfo,
                fNbytesInfo,
                fUUID_version,
                fUUID,
            ) = _file_header_fields_big.unpack(
                rawfile.read(_file_header_fields_big.size)
            )

        rawfile.seek(fSeekFree)
        freesegments = read_key(rawfile)

        rawfile.seek(freesegments["fSeekKey"] + freesegments["fKeylen"])

        frees = []
        for _ in range(nfree):
            version, fFirst, fLast = _free_format_small.unpack(
                rawfile.read(_free_format_small.size)
            )
            if version > 1000:
                rawfile.seek(rawfile.tell() - _free_format_small.size)
                version, fFirst, fLast = _free_format_big.unpack(
                    rawfile.read(_free_format_big.size)
                )
            frees.append((version, fFirst, fLast))

        directory_start = fBEGIN + fNbytesName

        rawfile.seek(directory_start)
        directory = read_directory(rawfile)

        rawfile.seek(directory["fSeekKeys"])

        beginning_of_scribble_1 = rawfile.tell()

        header_key = read_key(rawfile)
        (num_keys,) = _directory_format_num_keys.unpack(
            rawfile.read(_directory_format_num_keys.size)
        )
        keys = [read_key(rawfile) for _ in range(num_keys)]

        end_of_scribble_1 = rawfile.tell()

        beginning_of_scribble_2 = fSeekFree
        end_of_scribble_2 = fSeekFree + fNbytesFree

        # scribble over the directory and freesegments data

        rawfile.seek(beginning_of_scribble_1)
        rawfile.write(b"X" * (end_of_scribble_1 - beginning_of_scribble_1))

        rawfile.seek(beginning_of_scribble_2)
        rawfile.write(b"x" * (end_of_scribble_2 - beginning_of_scribble_2))

        # now make a new directory with the same contents

        _, beginning_of_new_directory, _ = frees[-1]

        rawfile.seek(beginning_of_new_directory)
        header_key["fSeekKey"] = beginning_of_new_directory

        write_key(rawfile, header_key)
        rawfile.write(_directory_format_num_keys.pack(num_keys))
        for key in keys:
            write_key(rawfile, key)

        end_of_new_directory = rawfile.tell()

        nfree = len(frees[:-1]) + 3
        fNbytesFree = (
            _key_format_small.size  # assuming small
            + length_string(freesegments["fClassName"])
            + length_string(freesegments["fName"])
            + length_string(freesegments["fTitle"])
            + nfree * _free_format_small.size  # assuming small
        )
        frees = frees[:-1] + [
            (
                1,
                beginning_of_scribble_1,
                end_of_scribble_1 - 1,
            ),  # fLast is really *last*
            (
                1,
                beginning_of_scribble_2,
                end_of_scribble_2 - 1,
            ),  # *not* the first unfree
            (1, end_of_new_directory + fNbytesFree, 2000000000),
        ]

        fSeekFree = rawfile.tell()

        freesegments["fSeekKey"] = fSeekFree
        freesegments["fNbytes"] = fNbytesFree
        freesegments["fObjlen"] = nfree * _free_format_small.size  # assuming small

        write_key(rawfile, freesegments)
        for free in frees:
            version, fFirst, fLast = free
            if version <= 1000:
                rawfile.write(_free_format_small.pack(version, fFirst, fLast))
            else:
                rawfile.write(_free_format_big.pack(version, fFirst, fLast))

        fEND = rawfile.tell()

        directory["fSeekKeys"] = beginning_of_new_directory

        rawfile.seek(directory_start)
        write_directory(rawfile, directory)

        rawfile.seek(0)
        if fVersion <= 1000000:
            rawfile.write(
                _file_header_fields_small.pack(
                    magic,
                    fVersion,
                    fBEGIN,
                    fEND,
                    fSeekFree,
                    fNbytesFree,
                    nfree,
                    fNbytesName,
                    fUnits,
                    fCompress,
                    fSeekInfo,
                    fNbytesInfo,
                    fUUID_version,
                    fUUID,
                )
            )
        else:
            rawfile.write(
                _file_header_fields_big.pack(
                    magic,
                    fVersion,
                    fBEGIN,
                    fEND,
                    fSeekFree,
                    fNbytesFree,
                    nfree,
                    fNbytesName,
                    fUnits,
                    fCompress,
                    fSeekInfo,
                    fNbytesInfo,
                    fUUID_version,
                    fUUID,
                )
            )

    # print("AFTER MY UPDATE")
    # for version, fFirst, fLast in get_frees(filename):
    #     print(version, fFirst, fLast)

    # read it back in ROOT and add another object

    f1 = ROOT.TFile(filename, "UPDATE")

    if source == "uproot-Zmumu.root":
        t = f1.Get("events")
        assert str(t.GetBranch("px1").GetTitle()) == "px1/D"
    elif source == "uproot-issue243-new.root":
        t = f1.Get("sig")
        assert str(t.GetBranch("m_hh").GetTitle()) == "m_hh/D"
    elif source == "uproot-issue64.root":
        t = f1.Get("events/events")
        assert str(t.GetBranch("eventid").GetTitle()) == "eventid/I"

    x = ROOT.TObjString("hello")
    x.Write()
    f1.Close()

    # print("AFTER ROOT'S UPDATE")
    # for version, fFirst, fLast in get_frees(filename):
    #     print(version, fFirst, fLast)

    f2 = ROOT.TFile(filename)
    assert str(f2.Get("hello")) == "hello"
    f2.Close()

    with uproot.open(filename) as f3:
        assert str(f3["hello"]) == "hello"
        key = f3.key("hello")
        # print(key.fSeekKey, key.fKeylen, key.fNbytes)


def read_string(rawfile):
    (string_size,) = _string_size_format_1.unpack(rawfile.read(1))
    if string_size == 255:
        (string_size,) = _string_size_format_4.unpack(rawfile.read(4))
    return rawfile.read(string_size).decode(errors="surrogateescape")


def length_string(string):
    bytestring = string.encode(errors="surrogateescape")
    length = len(bytestring)
    if len(bytestring) < 255:
        return len(bytestring) + 1
    else:
        return len(bytestring) + 5


def write_string(rawfile, string):
    bytestring = string.encode(errors="surrogateescape")
    length = len(bytestring)
    if len(bytestring) < 255:
        rawfile.write(struct.pack(">B%ds" % length, length, bytestring))
    else:
        rawfile.write(struct.pack(">BI%ds" % length, 255, length, bytestring))


def read_directory(rawfile):
    (
        fVersion,
        fDatimeC,
        fDatimeM,
        fNbytesKeys,
        fNbytesName,
        fSeekDir,
        fSeekParent,
        fSeekKeys,
    ) = _directory_format_small.unpack(rawfile.read(_directory_format_small.size))

    if fVersion > 1000:
        rawfile.seek(directory_start)
        (
            fVersion,
            fDatimeC,
            fDatimeM,
            fNbytesKeys,
            fNbytesName,
            fSeekDir,
            fSeekParent,
            fSeekKeys,
        ) = _directory_format_big.unpack(rawfile.read(_directory_format_big.size))

    return {
        "fVersion": fVersion,
        "fDatimeC": fDatimeC,
        "fDatimeM": fDatimeM,
        "fNbytesKeys": fNbytesKeys,
        "fNbytesName": fNbytesName,
        "fSeekDir": fSeekDir,
        "fSeekParent": fSeekParent,
        "fSeekKeys": fSeekKeys,
    }


def write_directory(rawfile, directory):
    if directory["fVersion"] <= 1000:
        rawfile.write(
            _directory_format_small.pack(
                directory["fVersion"],
                directory["fDatimeC"],
                directory["fDatimeM"],
                directory["fNbytesKeys"],
                directory["fNbytesName"],
                directory["fSeekDir"],
                directory["fSeekParent"],
                directory["fSeekKeys"],
            )
        )
    else:
        rawfile.write(
            _directory_format_big.pack(
                directory["fVersion"],
                directory["fDatimeC"],
                directory["fDatimeM"],
                directory["fNbytesKeys"],
                directory["fNbytesName"],
                directory["fSeekDir"],
                directory["fSeekParent"],
                directory["fSeekKeys"],
            )
        )


def read_key(rawfile):
    position = rawfile.tell()

    (
        fNbytes,
        fVersion,
        fObjlen,
        fDatime,
        fKeylen,
        fCycle,
        fSeekKey,
        fSeekPdir,
    ) = _key_format_small.unpack(rawfile.read(_key_format_small.size))

    if fVersion > 1000:
        rawfile.seek(position)
        (
            fNbytes,
            fVersion,
            fObjlen,
            fDatime,
            fKeylen,
            fCycle,
            fSeekKey,
            fSeekPdir,
        ) = _key_format_big.unpack(rawfile.read(_key_format_big.size))

    fClassName = read_string(rawfile)
    fName = read_string(rawfile)
    fTitle = read_string(rawfile)

    return {
        "fNbytes": fNbytes,
        "fVersion": fVersion,
        "fObjlen": fObjlen,
        "fDatime": fDatime,
        "fKeylen": fKeylen,
        "fCycle": fCycle,
        "fSeekKey": fSeekKey,
        "fSeekPdir": fSeekPdir,
        "fClassName": fClassName,
        "fName": fName,
        "fTitle": fTitle,
    }


def write_key(rawfile, key):
    if key["fVersion"] <= 1000:
        rawfile.write(
            _key_format_small.pack(
                key["fNbytes"],
                key["fVersion"],
                key["fObjlen"],
                key["fDatime"],
                key["fKeylen"],
                key["fCycle"],
                key["fSeekKey"],
                key["fSeekPdir"],
            )
        )
    else:
        rawfile.write(
            _key_format_big.pack(
                key["fNbytes"],
                key["fVersion"],
                key["fObjlen"],
                key["fDatime"],
                key["fKeylen"],
                key["fCycle"],
                key["fSeekKey"],
                key["fSeekPdir"],
            )
        )

    write_string(rawfile, key["fClassName"])
    write_string(rawfile, key["fName"])
    write_string(rawfile, key["fTitle"])


def get_frees(filename):
    with open(filename, "rb") as rawfile:
        rawfile.seek(0)
        (
            magic,
            fVersion,
            fBEGIN,
            fEND,
            fSeekFree,
            fNbytesFree,
            nfree,
            fNbytesName,
            fUnits,
            fCompress,
            fSeekInfo,
            fNbytesInfo,
            fUUID_version,
            fUUID,
        ) = _file_header_fields_small.unpack(
            rawfile.read(_file_header_fields_small.size)
        )

        if fVersion > 1000000:
            rawfile.seek(0)
            (
                magic,
                fVersion,
                fBEGIN,
                fEND,
                fSeekFree,
                fNbytesFree,
                nfree,
                fNbytesName,
                fUnits,
                fCompress,
                fSeekInfo,
                fNbytesInfo,
                fUUID_version,
                fUUID,
            ) = _file_header_fields_big.unpack(
                rawfile.read(_file_header_fields_big.size)
            )

        rawfile.seek(fSeekFree)
        freesegments = read_key(rawfile)

        rawfile.seek(freesegments["fSeekKey"] + freesegments["fKeylen"])

        frees = []
        for _ in range(nfree):
            version, fFirst, fLast = _free_format_small.unpack(
                rawfile.read(_free_format_small.size)
            )
            if version > 1000:
                rawfile.seek(rawfile.tell() - _free_format_small.size)
                version, fFirst, fLast = _free_format_big.unpack(
                    rawfile.read(_free_format_big.size)
                )
            frees.append((version, fFirst, fLast))

        return frees



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines a versionless model for ``TClonesArray``.
"""


from collections.abc import Sequence

import uproot.models.TObjArray


class Model_TClonesArray(uproot.model.VersionedModel, Sequence):
    """
    A versionless :doc:`uproot.model.Model` for ``TObjArray``.

    This also satisfies Python's abstract ``Sequence`` protocol.
    """

    def read_members(self, chunk, cursor, context, file):
        context["cancel_forth"] = True
        if self.is_memberwise:
            raise NotImplementedError(
                "memberwise serialization of {}\nin file {}".format(
                    type(self).__name__, self.file.file_path
                )
            )

        self._bases.append(
            uproot.models.TObject.Model_TObject.read(
                chunk,
                cursor,
                context,
                file,
                self._file,
                self._parent,
                concrete=self.concrete,
            )
        )

        self._members["fName"] = cursor.string(chunk, context)

        classname_and_version = cursor.string(chunk, context)
        try:
            i = classname_and_version.index(";")
        except ValueError:
            self._item_classname = classname_and_version
            self._item_classversion = "max"
        else:
            self._item_classname = classname_and_version[:i]
            self._item_classversion = int(classname_and_version[i + 1 :])

        self._members["fClass"] = cls = file.class_named(
            self._item_classname, self._item_classversion
        )

        self._members["fSize"], self._members["fLowerBound"] = cursor.fields(
            chunk, uproot.models.TObjArray._tobjarray_format1, context
        )

        self._data = []
        for _ in range(self._members["fSize"]):
            if cursor.byte(chunk, context) != 0:
                self._data.append(
                    cls.read(chunk, cursor, context, file, self._file, self._parent)
                )

    @property
    def item_classname(self):
        return self._item_classname

    @property
    def item_classversion(self):
        return self._item_classversion

    @property
    def item_class(self):
        return self._members["fClass"]

    def __getitem__(self, where):
        return self._data[where]

    def __len__(self):
        return len(self._data)

    base_names_versions = [("TObjArray", 3)]
    member_names = []
    class_flags = {}


uproot.classes["TClonesArray"] = Model_TClonesArray



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines a physical layer for file-like objects.

Defines a :doc:`uproot.source.object.ObjectResource` (wrapped Python file-like
object) and one source :doc:`uproot.source.object.ObjectSource` which always
has exactly one worker (we can't assume that the object is thread-safe).
"""


import uproot
import uproot.source.chunk
import uproot.source.futures


class ObjectResource(uproot.source.chunk.Resource):
    """
    Args:
        obj: The file-like object to use.

    A :doc:`uproot.source.chunk.Resource` for a file-like object.

    This object must have the following methods:

    - ``read(num_bytes)`` where ``num_bytes`` is an integer number of bytes to
      read.
    - ``seek(position)`` where ``position`` is an integer position to seek to.

    Both of these methods change the internal state of the object, its current
    seek position (because ``read`` moves that position forward ``num_bytes``).
    Hence, it is in principle not thread-safe.
    """

    def __init__(self, obj):
        self._obj = obj

    @property
    def obj(self):
        return self._obj

    @property
    def closed(self):
        return getattr(self._obj, "closed", False)

    def __enter__(self):
        if hasattr(self._obj, "__enter__"):
            self._obj.__enter__()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if hasattr(self._obj, "__exit__"):
            self._obj.__exit__(exception_type, exception_value, traceback)

    def get(self, start, stop):
        """
        Args:
            start (int): Seek position of the first byte to include.
            stop (int): Seek position of the first byte to exclude
                (one greater than the last byte to include).

        Returns a Python buffer of data between ``start`` and ``stop``.
        """
        self._obj.seek(start)
        return self._obj.read(stop - start)

    @staticmethod
    def future(source, start, stop):
        """
        Args:
            source (:doc:`uproot.source.object.ObjectSource`): The data source.
            start (int): Seek position of the first byte to include.
            stop (int): Seek position of the first byte to exclude
                (one greater than the last byte to include).

        Returns a :doc:`uproot.source.futures.ResourceFuture` that calls
        :ref:`uproot.source.object.ObjectResource.get` with ``start`` and
        ``stop``.
        """

        def task(resource):
            return resource.get(start, stop)

        return uproot.source.futures.ResourceFuture(task)


class ObjectSource(uproot.source.chunk.MultithreadedSource):
    """
    Args:
        obj: The file-like object to use.

    A :doc:`uproot.source.chunk.Source` for a file-like object. (Although this
    is a :doc:`uproot.source.chunk.MultithreadedSource`, it never has more or
    less than one thread.)

    This object must have the following methods:

    - ``read(num_bytes)`` where ``num_bytes`` is an integer number of bytes to
      read.
    - ``seek(position)`` where ``position`` is an integer position to seek to.

    Both of these methods change the internal state of the object, its current
    seek position (because ``read`` moves that position forward ``num_bytes``).
    Hence, it is in principle not thread-safe.
    """

    ResourceClass = ObjectResource

    def __init__(self, obj, **options):
        self._num_requests = 0
        self._num_requested_chunks = 0
        self._num_requested_bytes = 0

        self._file_path = repr(obj)
        self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
            [ObjectResource(obj)]
        )
        self._num_bytes = None
